"""Module for constrained DFT

for review see Chem. Rev., 2012, 112 (1), pp 321-370
article on GPAW implementation:
J. Chem. Theory Comput., 2016, 12 (11), pp 5367-5378
"""

import copy
import functools
from math import pi

import numpy as np
from ase.calculators.calculator import Calculator
from ase.data import atomic_numbers, covalent_radii
from ase.units import Bohr, Hartree
from ase.utils import IOContext
from scipy.optimize import minimize

from gpaw.external import ExternalPotential


class CDFT(Calculator):
    implemented_properties = ['energy', 'forces']

    def __init__(self,
                 calc,
                 atoms,
                 charge_regions=[],
                 charges=None,
                 spin_regions=[],
                 spins=None,
                 charge_coefs=None,
                 spin_coefs=None,
                 promolecular_constraint=False,
                 txt='-',
                 minimizer_options={'gtol': 0.01},
                 Rc={},
                 mu={'Li': 0.5,
                     'F': 0.7,
                     'O': 0.7,
                     'V': 0.5},
                 method='BFGS',
                 forces='analytical',
                 use_charge_difference=False,
                 compute_forces=True,
                 maxstep=100,
                 tol=1e-3,
                 bounds=None,
                 restart=False,
                 hess=None):
        """Constrained DFT calculator.

        calc: GPAW instance
            DFT calculator object to be constrained.
        charge_regions: list of list of int
            Atom indices of atoms in the different charge_regions.
        spin_regions: list of list of int
            Atom indices of atoms in the different spin_regions.
        charges: list of float
            constrained charges in the different charge_regions.
        spins: list of float
            constrained spins in the different charge_regions.
            Value of 1 sets net magnetisation of one up/alpha electron
        charge_coefs: list of float
            Initial values for charge constraint coefficients (eV).
        spin_coefs: list of float
            Initial values for spin constraint coefficients (eV).
        promolecular_constraint: bool
            Define charge and/or spin constraints from promolecular
            densities, see: dx.doi.org/10.1021/cr200148b Eq. 29-31
            If true, user specified charges/spins are overwritten!
            The atoms (of Atoms object) specifying the charge/spin regions
            need to contain have correct charge/spin state!
            (atoms.set_initial_charges([atomic_charges]) and
            atoms.set_initial_magnetic_moments([moments]))
        txt: None or str or file descriptor
            Log file.  Default id '-' meaning standard out.  Use None for
            no output.
        minimizer_options: dict
            options for scipy optimizers, see:scipy.optimize.minimize
        method: str
            One of scipy optimizers, e.g., BFGS, CG
        forces: str
            cDFT weight function contribution to forces
            'fd' for finite difference or 'analytical'
        difference: bool
            If True, constrain charge difference between two regions
            Then charge_regions needs two regions and charges needs
            only one item which is the charge difference between
            the two regions, the first beign donor, the second acceptor

            If False, each region is treated with the corresponding
            charge constraint
        compute_forces: bool
            Should the forces be computed?
        restart: bool
            starting from an old calculation from gpw
        hess: '2-point', '3-point', 'cs'
            scipy hessian approximation
        """

        Calculator.__init__(self)

        self.calc = calc
        self.restart = restart
        self.iocontext = IOContext()
        self.log = self.iocontext.openfile(txt, calc.world)
        self.method = method
        self.forces = forces
        self.options = minimizer_options
        self.difference = use_charge_difference
        self.compute_forces = compute_forces
        self.Rc = Rc
        self.mu = mu
        # set charge constraints and lagrangians
        self.v_i = np.empty(shape=(0, 0))

        self.constraints = np.empty(shape=(0, 0))

        self.charge_regions = charge_regions
        self.spin_regions = spin_regions
        self._n_charge_regions = len(self.charge_regions)
        self._n_spin_regions = len(self.spin_regions)
        self._n_regions = self._n_charge_regions + self._n_spin_regions

        self.max_step = maxstep
        self.tol = tol
        self.gtol = minimizer_options['gtol']
        self.bounds = bounds

        if self.bounds is not None:
            self.bounds = np.asarray(self.bounds) / Hartree
        self.hess = hess

        if self.difference:
            # difference calculation only for 2 charge regions
            if self.n_spin_regions != 0 or self._n_charge_regions != 2:
                raise ValueError('No spin constraints '
                                 'for charge difference calculations and'
                                 ' only two regions allowed')
            assert (self.n_charge_regions == 1)

        if self.n_charge_regions == 0:
            self.regions = []

        else:
            self.charge_i = np.array(charges, dtype=float)

            if charge_coefs is None:  # to Hartree
                self.v_i = 0.1 * np.sign(self.charge_i)
            else:
                self.v_i = np.array(charge_coefs) / Hartree

            if not self.difference:
                self.regions = copy.copy(self.charge_regions)

                # The objective is to constrain the number of electrons (nel)
                # in a certain region --> convert charge to nel
                Zn = np.zeros(self.n_charge_regions)
                for j in range(len(Zn)):
                    for atom in atoms[self.charge_regions[j]]:
                        Zn[j] += atom.number

                # combined spin and charge constraints
                self.constraints = Zn - self.charge_i

            else:  # constrain charge between two regions
                nD = 0.  # neutral donor
                nA = 0.  # neutral acceptor

                for atom in atoms[charge_regions[0]]:
                    nD += atom.number
                for atom in atoms[charge_regions[1]]:
                    nA += atom.number

                self.dn_core = nD - nA  # difference of core
                self.constraints = [self.dn_core - charges[0]]
                self.regions = charge_regions

        # set spin constraints
        if self.n_spin_regions != 0 and not self.difference:
            spin_i = np.array(spins, dtype=float)
            self.constraints = np.append(self.constraints, spin_i)

            if spin_coefs is None:  # to Hartree
                v_is = 0.1 * np.sign(spin_i)
            else:
                v_is = np.array(spin_coefs) / Hartree

            self.v_i = np.append(self.v_i, v_is)
            # self.regions.append(spin_regions)
            for i in range(self.n_spin_regions):
                self.regions.append(self.spin_regions[i])

        # initialise without v_ext
        atoms.calc = self.calc
        if not self.restart:
            atoms.get_potential_energy()

        assert atoms.calc.wfs.nspins == 2

        self.cdft_initialised = False

        self.atoms = atoms
        self.gd = self.calc.density.finegd

        if promolecular_constraint:
            self.constraints = get_promolecular_constraints(
                calc=self.calc,
                atoms=self.atoms,
                charge_regions=self.charge_regions,
                spin_regions=self.spin_regions,
                charges=charges,
                spins=spins)

        # get number of core electrons at each constrained region
        # used for pseudo free energy to neglect core contributions
        # in coupling calculation
        self.n_core_electrons = np.zeros((len(self.regions)))
        for a in self.atoms:
            for r in range(len(self.regions[:self.n_charge_regions])):
                if a.index in self.regions[r] and not self.difference:
                    n_core = a.number - self.calc.wfs.setups[a.index].Nv
                    self.n_core_electrons[r] += n_core
                elif a.index in self.regions[r] and self.difference:
                    if r == 0:
                        n_core = a.number - self.calc.wfs.setups[a.index].Nv
                        self.n_core_electrons[r] += n_core
                    else:
                        n_core = a.number - self.calc.wfs.setups[a.index].Nv
                        self.n_core_electrons[r] -= n_core

        w = WeightFunc(self.gd, self.atoms, None, self.Rc, self.mu)
        self.Rc, self.mu = w.get_Rc_and_mu()
        # construct cdft potential
        self.ext = CDFTPotential(regions=self.regions,
                                 gd=self.gd,
                                 atoms=self.atoms,
                                 constraints=self.constraints,
                                 n_charge_regions=self.n_charge_regions,
                                 difference=self.difference,
                                 txt=self.log,
                                 vi=self.v_i,
                                 Rc=self.Rc,
                                 mu=self.mu)

        self.calc.set(external=self.ext)

        self.w = self.ext.w_ig

    def __del__(self):
        self.iocontext.close()

    def calculate(self, atoms, properties, system_changes):
        # check we're dealing with same atoms
        if atoms != self.atoms:
            self.atoms = atoms
        if not self.restart:
            Calculator.calculate(self, self.atoms)

        Calculator.calculate(self, self.atoms)

        # update positions and weight functions
        if 'positions' in system_changes or not self.cdft_initialised:
            self.ext.set_positions_and_atomic_numbers(
                self.atoms.positions / Bohr, self.atoms.numbers)
            self.cdft_initialised = True

        self.atoms.calc = self.calc

        p = functools.partial(print, file=self.log)
        self.iteration = 0
        self.old_v_i = self.v_i.copy()

        def f(v_i):
            # nonlocal iteration
            # very simple step size control
            diff = v_i - self.old_v_i

            if np.any(np.abs(diff) >= self.max_step / Hartree):
                v_i = self.old_v_i + np.sign(
                    v_i - self.old_v_i) * self.max_step / Hartree

            self.ext.set_levels(v_i)
            self.v_i = v_i
            # cDFT free energy <A|H^KS + V_a w_a|A> = Edft + <A|Vw|A>
            self.Ecdft = self.atoms.get_potential_energy()  # in eV

            # cDFT corrections to energy
            self.get_atomic_density_correction()
            self.Ecdft += self.get_energy_correction() * Hartree

            # get the cDFT gradient
            dn_i = []
            Delta_n = self.get_energy_correction(return_density=True)

            if self.calc.density.nt_sg is None:
                self.density.interpolate_pseudo_density()

            self.nt_ag = self.calc.density.nt_sg[0]
            self.nt_bg = self.calc.density.nt_sg[1]
            total_electrons = []

            if self.n_charge_regions != 0:
                # total pseudo electron density
                n_gt = self.nt_ag + self.nt_bg

                n_electrons = (self.gd.integrate(
                    self.ext.w_ig[0:self.n_charge_regions] * n_gt,
                    global_integral=True))

                # corrections
                n_electrons += Delta_n[0:self.n_charge_regions]

                # constraint
                diff = n_electrons - self.constraints[0:self.n_charge_regions]
                total_electrons.append(n_electrons)
                dn_i.append(diff)

            if self.n_spin_regions != 0:
                # difference of pseudo spin densities
                Dns_gt = (self.nt_ag - self.nt_bg)
                n_electrons = self.gd.integrate(
                    self.ext.w_ig[self.n_charge_regions:] * Dns_gt,
                    global_integral=True)
                # corrections
                n_electrons += Delta_n[self.n_charge_regions:]
                # constraint
                diff = n_electrons - self.constraints[self.n_charge_regions:]
                total_electrons.append(n_electrons)
                dn_i.append(diff)

            self.dn_i = np.asarray(dn_i)
            self.dn_i = self.dn_i.flatten()
            total_electrons = np.asarray(total_electrons)
            self.w = self.ext.w_ig
            # Do not include external potential
            E_KS = get_ks_energy_wo_external(self.calc)

            self.Edft = E_KS
            # pseudo free energy, neglecting core electrons as done for
            # coupling constant calculation
            if not self.difference:

                self.Ecdft = E_KS + np.dot(
                    self.v_i * Hartree,
                    (total_electrons - self.n_core_electrons)[0])

            if self.iteration == 0:
                p('Optimizer: {n}'.format(n=self.method))
                p('Optimizer setups:{n}'.format(n=self.options))

            header = '----------------------------------------\n'
            header += 'Iteration: {0}\n'
            header += 'Energy: {1:10.8f} eV\n'
            header += 'Coefficients [eV]: {2} \n'
            header += 'Errors [e]: {3} \n'

            dn_intermediate = self.dn_i.tolist()
            p(
                header.format(self.iteration,
                              self.Edft,
                              ''.join('{0:4.3f}  '.format(v)
                                      for v in self.v_i * Hartree),
                              ''.join('{0:6.4f}  '.format(dn)
                                      for dn in dn_intermediate),
                              fill='left',
                              align='left'))

            self.log.flush()
            self.iteration += 1

            self.old_v_i = self.v_i.copy()
            return np.max(np.abs(self.dn_i))

        m = minimize(f,
                     self.v_i,
                     jac=self.jacobian,
                     hess=self.hess,
                     bounds=self.bounds,
                     tol=self.tol,
                     method=self.method,
                     options=self.options)

        p(str(m.message) + '\n')

        self.density = self.calc.density  # TS09-vdw needs this
        self.results['energy'] = self.Edft

        # print to log

        p('Final DFT energy : ' + str(self.Edft) + ' eV \n')
        p('CDFT free energy <A|H+Vw|A> : ' + str(self.Ecdft) + ' eV \n')

        if self.compute_forces:
            f = WeightFunc(self.gd,
                           self.atoms,
                           self.regions,
                           self.Rc,
                           self.mu,
                           new=False)

            f_cdft = f.get_cdft_forces2(dens=self.calc.density,
                                        v_i=self.v_i,
                                        n_charge_regions=self.n_charge_regions,
                                        n_spin_regions=self.n_spin_regions,
                                        w_ig=self.w,
                                        method=self.forces,
                                        difference=self.difference)

            self.calc.wfs.world.broadcast(f_cdft, 0)
            self.ext.set_forces(f_cdft)
            self.results['forces'] = self.atoms.get_forces()

    def get_weight(self, save=True, name='weight', pad=False):
        if not pad:
            w_g = self.w
        else:
            c = CDFTPotential(regions=self.regions,
                              gd=self.gd,
                              atoms=self.atoms,
                              constraints=self.constraints,
                              n_charge_regions=self.n_charge_regions,
                              difference=self.difference,
                              vi=self.v_i,
                              Rc=self.Rc,
                              mu=self.mu)

            w_g = c.initialize_partitioning(self.gd,
                                            construct=True,
                                            pad=True,
                                            global_array=True)
        if save:
            w_s = self.gd.collect(w_g, broadcast=True)

            if self.gd.comm.rank == 0:
                np.save('coarse_weight', w_s)
        return w_g

    def jacobian(self, v_i):
        if np.all((np.abs(self.dn_i) < self.gtol)):
            # forces scipy opt to converge when gtol is reached
            return np.zeros(len(self.v_i))
        else:
            return -self.dn_i
    
    def cdft_free_energy(self):
        return self.Ecdft

    def dft_energy(self):
        return self.Edft

    def get_lagrangians(self):
        return self.v_i * Hartree

    def get_constraints(self):
        return self.constraints

    def get_grid(self):
        return self.gd

    def get_all_electron_density(self, gridrefinement=2, spin=None):
        return self.calc.get_all_electron_density(
            gridrefinement=gridrefinement, spin=spin)

    def get_atomic_density_correction(self, return_els=False):
        # eq. 20 of the paper
        self.dn_s = np.zeros((2, len(self.atoms)))

        for i in [0, 1]:
            for a, D_sp in self.calc.density.D_asp.items():
                self.dn_s[i, a] += np.sqrt(4 * np.pi) * (
                    np.dot(D_sp[i],
                           self.calc.wfs.setups[a].Delta_pL)[0] +
                    self.calc.wfs.setups[a].Delta0 / 2)

        self.gd.comm.sum(self.dn_s)
        for a in range(len(self.atoms)):
            self.dn_s[:, a] += self.atoms[a].number / 2.
        if return_els:
            return self.dn_s

    def get_energy_correction(self, return_density=False):
        # Delta n^a part of eq 21

        # for each region
        n_a = np.zeros(len(self.regions))

        # int w_i Dn_i for both spins
        # in spin constraints w_ib = -w_ia
        # inside augmentation spheres w_i = 1

        for c in range(len(self.regions)):
            # sum all atoms in a region
            n_sa = self.dn_s[0, self.regions[c]].sum()
            n_sb = self.dn_s[1, self.regions[c]].sum()
            # total density correction
            n_a[c] = n_sa + n_sb

        for s in range(self.n_spin_regions):
            n_sa = self.dn_s[0, self.regions[self.n_charge_regions + s]].sum()
            n_sb = self.dn_s[1, self.regions[self.n_charge_regions + s]].sum()
            n_a[self.n_charge_regions + s] = n_sa - n_sb

        if return_density:
            if not self.difference:
                # Delta n^a, eq 20
                return n_a
            else:
                # the difference of corrections
                return [n_a[0] - n_a[1]]

        else:
            if not self.difference:
                return (np.dot(self.v_i, n_a))
            else:
                # negative for difference acceptor
                vi_temp = np.array([self.v_i[0], -self.v_i[0]])
                return (np.dot(vi_temp, n_a))

    def get_number_of_electrons_on_atoms(self):
        """Return the number of electrons with each gaussian."""

        nelectrons = []
        ae_dens_correction = self.get_atomic_density_correction(
            return_els=True)
        if self.calc.density.nt_sg is None:
            self.calc.density.interpolate_pseudo_density()

        dens = self.calc.density.nt_sg[0] + self.calc.density.nt_sg[1]

        for atom in self.atoms:
            # weight function with one atom
            f = WeightFunc(self.gd,
                           self.atoms, [atom.index],
                           self.Rc,
                           self.mu,
                           new=False)

            w = f.construct_weight_function()
            n_el = (self.gd.integrate(w * dens, global_integral=True))
            # corrections
            n_el += (ae_dens_correction[:, atom.index]).sum()
            nelectrons.append(n_el)

        return nelectrons

    def write(self, name, mode=None):
        self.calc.write(name, mode=mode)

    def save_parameters(self, name='initial', save_weight=True, save_wfs=True):
        if self.gd.comm.rank == 0:
            file = open(name + '.txt', 'w')
            file.write('NA = %f ,\n' % (self.constraints))
            file.write('FA = %f , \n' % (self.Ecdft))
            file.write('EA = %f , \n' % (self.Edft))
            file.write('Va = %f , \n' % (self.v_i * Hartree))
            file.write('N_charge_regions_A = %d ,\n' % self.n_charge_regions)
            file.close()

    def get_weight_function_on_coarse_grid(self,
                                           regions,
                                           gd,
                                           atoms,
                                           constraints,
                                           n_charge_regions,
                                           difference,
                                           save=True):

        gd = self.calc.density.gd
        c = CDFTPotential(regions=self.regions,
                          gd=gd,
                          atoms=self.atoms,
                          constraints=self.constraints,
                          n_charge_regions=self.n_charge_regions,
                          difference=self.difference,
                          vi=self.v_i)

        w_G = c.initialize_partitioning(gd, construct=True)

        if save:
            w_s = gd.collect(w_G, broadcast=True)
            if gd.comm.rank == 0:
                np.save('coarse_weight', w_s)
        return w_G

    def get_coarse_grid(self, save=True):

        if save:
            w_s = self.calc.density.gd.collect(self.gd, broadcast=True)
            if self.calc.density.gd.comm.rank == 0:
                np.save('coarse_grid', w_s)

        return self.calc.density.gd

    @property
    def n_spin_regions(self):
        self._n_spin_regions = len(self.spin_regions)
        return self._n_spin_regions

    @property
    def n_charge_regions(self):
        if not self.difference:
            self._n_charge_regions = len(self.charge_regions)
        else:
            self._n_charge_regions = 1
        return self._n_charge_regions

    @property
    def n_regions(self):
        if not self.difference:
            self._n_regions = self._n_charge_regions + self._n_spin_regions
        else:
            self._n_regions = 2
        return self._n_regions


def gaussians(gd, positions, numbers):
    r_Rv = gd.get_grid_point_coordinates().transpose((1, 2, 3, 0))
    radii = covalent_radii[numbers]
    cutoffs = radii + 3.0
    sigmas = radii * min(covalent_radii) + 0.5
    result_R = gd.zeros()
    for pos, Z, rc, sigma in zip(positions, numbers, cutoffs, sigmas):
        d2_R = ((r_Rv - pos)**2).sum(3)
        a_R = Z / (sigma**3 * (2 * np.pi)**1.5) * np.exp(-d2_R /
                                                         (2 * sigma**2))
        a_R[d2_R > rc] = 0.0
        result_R += a_R
    return result_R


class CDFTPotential(ExternalPotential):
    def __init__(self,
                 regions,
                 gd,
                 atoms,
                 constraints,
                 n_charge_regions,
                 difference,
                 vi,
                 txt='-',
                 Rc={},
                 mu={}):

        self.indices_i = regions
        self.gd = gd
        self.iocontext = IOContext()
        self.log = self.iocontext.openfile(txt)
        self.atoms = atoms
        self.pos_av = None
        self.Z_a = None
        self.w_ig = None
        self.n_charge_regions = n_charge_regions
        self.constraints = constraints
        self.difference = difference
        self.v_i = vi
        self.Rc = Rc
        self.mu = mu
        self.name = 'CDFTPotential'

    def __del__(self):
        self.iocontext.close()

    def __str__(self):
        self.name = 'CDFTPotential'
        return 'CDFTPotential'

    def get_name(self):
        return self.name

    def update_ae_density(self, ae_dens):
        self.ae_dens = ae_dens

    def get_atoms(self):
        return self.atoms

    def get_ae_density(self):
        return self.ae_dens

    def get_vi(self):
        return self.v_i

    def get_constraints(self):
        return self.constraints

    def set_levels(self, v_i):
        self.v_i = np.array(v_i, dtype=float)
        self.vext_g = None
        return self.v_i

    def set_forces(self, cdft_forces):
        self.cdft_forces = cdft_forces

    def get_cdft_forces(self):
        return self.cdft_forces

    def spin_polarized_potential(self):
        return len(self.constraints) != self.n_charge_regions

    def get_w(self):
        return self.w_ig

    def set_positions_and_atomic_numbers(self, pos_av, Z_a):
        self.pos_av = pos_av
        self.Z_a = Z_a
        self.w_ig = None
        self.vext_g = None

    def initialize_partitioning(self,
                                gd,
                                construct=False,
                                pad=False,
                                global_array=False):
        w = get_all_weight_functions(self.atoms, gd, self.indices_i,
                                     self.difference, self.Rc, self.mu)
        if construct:
            return np.array(w)

        self.w_ig = np.array(w)
        p = functools.partial(print, file=self.log)
        p('Number of charge constrained regions: {n}'.format(
            n=self.n_charge_regions))
        if not self.difference:
            p('Number of spin constrained regions: {n}'.format(
                n=len(self.indices_i) - self.n_charge_regions))
        else:
            p('Number of spin constrained regions: 0')
        p('Charge difference: {n}'.format(n=self.difference))
        p('Parameters')
        p('Atom      Width[A]      Rc[A]')
        for a in self.mu:
            p('  {atom}       {width:.3f}        {Rc:.3f}'.format(
                atom=a, width=self.mu[a] * Bohr, Rc=self.Rc[a] * Bohr))
        print(file=self.log)
        self.log.flush()

    def calculate_potential(self, gd):
        # return v_ext^{\sigma} = sum_i V_i*w_i^{\sigma}
        if self.w_ig is None:
            self.initialize_partitioning(self.gd)
        pot = np.empty(self.w_ig.shape)

        for i in range(len(self.v_i)):
            pot[i] = self.v_i[i] * self.w_ig[i]

        # first alpha spin
        vext_sga = np.sum(np.asarray(pot), axis=0)
        # then beta
        vext_sgb = np.asarray(pot)
        # spin constraints with beta spins
        vext_sgb[self.n_charge_regions:] *= -1.
        vext_sgb = np.sum(vext_sgb, axis=0)
        vext_sg = np.array([vext_sga, vext_sgb])
        # spin-dependent cdft potential
        self.vext_g = vext_sg

    def write(self, writer):
        writer.write(vext='CDFTPotential')

    def read(self, reader):
        pass

    def todict(self):
        return {
            'name': 'CDFTPotential',
            'regions': self.indices_i,
            'constraints': self.v_i * Hartree,
            'n_charge_regions': self.n_charge_regions,
            'difference': self.difference
        }

    def get_atomic_hamiltonians(self, setups, atom):
        h_cdft_a = np.zeros(setups.shape)
        h_cdft_b = np.zeros(setups.shape)
        v_i = self.v_i
        if self.difference:
            v_i = [v_i, -v_i]

        for i in range(len(self.indices_i)):
            if atom in self.indices_i[i]:
                h_cdft_a += v_i[i] * 2. * np.sqrt(np.pi) * setups
                h_cdft_b += v_i[i] * 2. * np.sqrt(np.pi) * setups
                if i >= self.n_charge_regions and not self.difference:
                    h_cdft_b *= -1.

        return h_cdft_a, h_cdft_b

    def get_cdft_external_energy(self, dens, nspins, vext_g, vt_g, vbar_g,
                                 vt_sg):
        # cDFT works with all-electron (spin)density

        # First the potential
        # spin dependent external potential, array of Vi*wi
        # vext_g = self.calculate_potential(self.gd).copy()
        vt_g += vext_g[0]
        vext_g[1:nspins] = vext_g[1] + vbar_g
        vt_sg[nspins:] = 0.0

        # then energy
        w = self.get_w()  # weight functions
        Vi = self.get_vi()
        constraints = self.get_constraints()

        # pseudo electron density on fine grid
        if dens.nt_sg is None:
            dens.interpolate_pseudo_density()

        diff = []

        if self.n_charge_regions != 0:
            # pseudo density
            nt_g = dens.nt_sg[0] + dens.nt_sg[1]
            charge_diff = (
                self.gd.integrate(w[0:self.n_charge_regions],
                                  nt_g, global_integral=True) -
                constraints[0:self.n_charge_regions])
            diff.append(charge_diff)

        # constrained spins
        if len(constraints) - self.n_charge_regions != 0:
            Delta_nt_g = dens.nt_sg[0] - dens.nt_sg[
                1]  # pseudo spin difference density
            spin_diff = self.gd.integrate(
                w[self.n_charge_regions:], Delta_nt_g,
                global_integral=True) - constraints[self.n_charge_regions:]
            diff.append(spin_diff)

        diff = np.asarray(diff)
        # number of domains
        size = self.gd.comm.size
        e = np.multiply(Vi, diff[:]).sum()
        e /= size

        return e


class WeightFunc:
    """ Class which builds a weight function around atoms or molecules
    given the atom index - using normalized Gaussian with cut-off!

    The weight function is constructed on the coarse or fine grid and
    can be used to do charge constraint DFT.

    """
    def __init__(self, gd, atoms, indices, Rc={}, mu={}, new=True):
        """ Given a grid-descriptor, atoms object and an index list
            construct a weight function defined by:
                     n_i(r-R_i)
            w(r) = ---------------
                   sum_a n_a(r-R_a)

            where a runs over all atoms, and i can index
            an atom or a list of atoms comprising a molecule, etc.

            The n_i are construced with atom centered gaussians
            using a pre-defined cut-off Rc_i.

        """
        self.gd = gd
        self.atoms = atoms
        self.indices_i = indices  # Indices of constrained charge_regions

        # Weight function parameters in Bohr
        # Cutoffs
        if new:
            new_Rc = {}
            for a in self.atoms:
                if a.symbol in Rc:
                    new_Rc[a.symbol] = Rc[a.symbol] / Bohr
                else:
                    element_number = atomic_numbers[a.symbol]
                    cr = covalent_radii[element_number]
                    # Rc to roughly between 3. and 5.
                    new_Rc[a.symbol] = (cr + 2.5) / Bohr

            self.Rc = new_Rc
        else:
            self.Rc = Rc

        # Construct mu (width) dict
        # mu only sets the width and height so it's in angstrom
        if new:
            new_mu = {}
            for a in self.atoms:
                if a.symbol in mu:
                    new_mu[a.symbol] = mu[a.symbol] / Bohr
                else:
                    element_number = atomic_numbers[a.symbol]
                    cr = covalent_radii[element_number]
                    # mu to be roughly between 0.5 and 1.0 AA
                    cr = (cr * min(covalent_radii) + 0.5)
                    new_mu[a.symbol] = cr / Bohr

            # "Larger" atoms may need a bit more width
            self.mu = new_mu
        else:
            self.mu = mu

    def get_Rc_and_mu(self):
        return self.Rc, self.mu

    def normalized_gaussian(self, dis, mu, Rc):
        # Given mu - width, and Rc
        # a normalized gaussian is constructed
        # around some atom. This is
        # placed on the gd (grid) - and truncated
        # at a given cut-off value Rc. dis
        # are the distances from atom to grid points.
        """ Normalized gaussian is:
                      1
        g(r) = ---------------  e^{-(r-Ra)^2/(2mu^2)}
               mu^3*(2pi)^1.5

        for |r-Ra| <= Rc, 0 elsewhere

        """
        # Check function
        check = abs(dis) <= Rc

        # Make gaussian 3D Guassian
        gauss = 1 / (mu * (2 * pi)**(1 / 2)) * np.exp(-dis**2 / (2.0 * mu**2))

        # apply cut-off and return
        return np.array(gauss * check)

    def get_distance_vectors(self, pos, distance=True):
        xyz = self.gd.get_grid_point_distance_vectors(pos)
        if distance:
            # returns vector norm
            return np.sqrt((xyz**2).sum(axis=0))
        else:
            # gives raw vector
            return xyz

    def construct_total_density(self, atoms):
        # Add to empty grid
        dens = self.gd.zeros()

        for atom in atoms:
            charge = atom.number
            symbol = atom.symbol
            pos = atom.position / Bohr

            dis = self.get_distance_vectors(pos)

            dens += charge * self.normalized_gaussian(dis,
                                                      self.mu[symbol],
                                                      self.Rc[symbol])
        return dens

    def construct_weight_function(self):
        # Grab atomic / molecular density
        dens_n = self.construct_total_density(self.atoms[self.indices_i])
        # Grab total density
        dens = self.construct_total_density(self.atoms)
        # Check zero elements
        check = dens == 0
        # Add value to zeros ...
        dens += check * 1.0
        # make weight function
        return (dens_n / dens)

    def get_cdft_forces2(self, dens, v_i, n_charge_regions, n_spin_regions,
                         w_ig, method, difference):
        """ Calculate cDFT force as a sum
        dF/dRi = Fi(inside) + Fs(surf)
        due to cutoff (Rc) in gauss
                  / dw(r)
        Fi_a = -V | ------ n(r) dr
                  /  dR_a
        dw(r)
        ----  = sum of Gaussian functions...
        dR_a

        this is computed in get_dw_dRa
        dens = density
        Vc = cDFT constraint value
        method = 'fd' or 'analytical' for
              finite difference or analytical
              dw/dR
        """

        cdft_forces = np.zeros((len(self.atoms), 3))

        if dens.nt_sg is None:
            dens.interpolate_pseudo_density()

        rho_kd = self.construct_total_density(self.atoms)  # sum_k n_k
        # Check zero elements
        check = rho_kd == 0.
        # Add value to zeros for denominator...
        rho_kd += check * 1.0

        for a, atom in enumerate(self.atoms):
            wn_sg = self.gd.zeros()
            prefactor = self.get_derivative_prefactor(n_charge_regions,
                                                      n_spin_regions, w_ig,
                                                      v_i,
                                                      difference,
                                                      atom, rho_kd)

            # make extended array
            for c in range(n_charge_regions):
                wn_sg += (dens.nt_sg[0] + dens.nt_sg[1]) * prefactor[0]

            for s in range(n_spin_regions):
                wn_sg += (dens.nt_sg[0] - dens.nt_sg[1]) * prefactor[1]

            for i in [0, 1, 2]:
                if method == 'analytical':
                    dG_dRav = self.get_analytical_derivates(atom,
                                                            direction=i)
                elif method == 'fd':
                    dG_dRav = self.get_fd_derivatives(atom,
                                                      direction=i)

                cdft_forces[a][i] -= self.gd.integrate(wn_sg * dG_dRav,
                                                       global_integral=True)

        return cdft_forces

    def get_fd_derivatives(self, atom, direction, dx=1.e-4):

        dirs = [[dx, 0., 0.], [0., dx, 0.], [0., 0., dx]]
        charge = atom.number
        symbol = atom.symbol
        mu = self.mu[symbol]
        Rc = self.Rc[symbol]

        a_posx = atom.position / Bohr + dirs[direction]
        a_dis = self.get_distance_vectors(a_posx)
        Ga_posx = charge * self.normalized_gaussian(a_dis, mu, Rc)
        # move to -dx
        a_negx = atom.position / Bohr - dirs[direction]
        a_dis = self.get_distance_vectors(a_negx)
        Ga_negx = charge * self.normalized_gaussian(a_dis, mu, Rc)
        # dG/dx
        dGav = (Ga_posx - Ga_negx) / (2 * dx)

        return dGav

    def get_derivative_prefactor(self, n_charge_regions, n_spin_regions, w_ig,
                                 v_i, difference, atom, rho_kd):
        """Computes the dw/dRa array needed for derivatives/forces
        eq 31
        needed for lfc-derivative/integrals
        """
        wc = self.gd.zeros()
        ws = self.gd.zeros()

        for i in range(n_charge_regions):
            # build V_i [sum_k rho_k + sum_{j in i}rho_i]
            wi = -w_ig[i]

            if not difference:
                if atom.index in self.indices_i[i]:
                    wi += 1.
            else:
                if atom.index in self.indices_i[0]:
                    wi += 1.
                elif atom.index in self.indices_i[1]:
                    wi -= 1.

            wi *= v_i[i]
            wc += wi / rho_kd

        for i in range(n_spin_regions):
            # build V_i [sum_k rho_k + sum_{j in i}rho_i]
            wi = -w_ig[n_charge_regions + i]
            if atom.index in self.indices_i[n_charge_regions + i]:
                wi += 1.
            wi *= v_i[n_charge_regions + i]
            ws += wi / rho_kd

        return [wc, ws]

    def get_analytical_derivates(self, atom, direction):
        # equations 32,33,34

        a_pos = atom.position / Bohr
        a_symbol = atom.symbol
        a_charge = atom.number

        a_dis = self.get_distance_vectors(a_pos)
        rRa = -self.get_distance_vectors(a_pos, distance=False)
        dist_rRa = self.get_distance_vectors(a_pos, distance=True)
        check = dist_rRa == 0

        # Add value to zeros ...
        dist_rRa += check * 1.0
        # eq 33
        drRa_di = rRa[direction] / dist_rRa

        # Gaussian derivative eq 34

        G_a = a_charge * self.normalized_gaussian(a_dis,
                                                  self.mu[a_symbol],
                                                  self.Rc[a_symbol])

        # within cutoff or at surface ? --> heaviside
        # inside
        check_i = abs(a_dis) <= self.Rc[a_symbol]
        rRc = check_i * a_dis
        dGa_drRa = -rRc * G_a / (
            self.mu[a_symbol])**2  # (\Theta * (r-R_a) n_A) / \sigma^2

        # at surface
        surf = abs(abs(a_dis) - self.Rc[a_symbol])
        check_s = surf <= max(self.gd.get_grid_spacings())
        dGa_drRa += check_s * G_a  # \ sigma_{A\in i} n_A

        # eq 32
        return dGa_drRa * drRa_di


def get_ks_energy_wo_external(calc):
    return (calc.hamiltonian.e_kinetic + calc.hamiltonian.e_coulomb +
            calc.hamiltonian.e_zero + calc.hamiltonian.e_xc -
            calc.hamiltonian.e_entropy) * Hartree


def get_all_weight_functions(atoms,
                             gd,
                             indices_i,
                             difference,
                             Rc,
                             mu,
                             new=False,
                             return_Rc_mu=False):
    w = []
    for i in range(len(indices_i)):
        wf = WeightFunc(gd, atoms, indices_i[i], Rc, mu, new)
        weig = wf.construct_weight_function()

        if not difference:
            w.append(weig)
        else:  # for charge difference constraint
            if i == 0:
                w.append(weig)
            else:
                w[0] -= weig  # negative for acceptor
    if return_Rc_mu:
        Rc, mu = wf.get_Rc_and_mu()
        return w, Rc, mu
    else:
        return w


def get_promolecular_constraints(calc_a,
                                 atoms_a,
                                 calc_b,
                                 atoms_b,
                                 charge_constraint=True,
                                 spin_constraint=False,
                                 restart=False,
                                 Rc={},
                                 mu={}):
    """
    - calc_a is for the region you're interested in. Its charge
        should correspond to the "free charge" of the promolecule
    - atoms_a: atoms object for the region of interest
    - calc_b is for the other region
    - atoms_b: atoms object for the other region
    - restart: bool. If true, atoms_a have used calc_a and atoms_b calc_b
    """

    constraints = []
    atoms = atoms_a + atoms_b

    if not restart:
        atoms_a.calc = calc_a
        atoms_b.calc = calc_b
    atoms_a.get_potential_energy()
    atoms_b.get_potential_energy()

    gd = calc_a.density.finegd

    if charge_constraint:
        # can charge and spin constraints be treated at the same time?
        d_a = calc_a.get_all_electron_density(gridrefinement=2, pad=False)
        d_b = calc_b.get_all_electron_density(gridrefinement=2, pad=False)

        charge_region = [atom.index for atom in atoms_a]
        weight = WeightFunc(gd=gd,
                            atoms=atoms,
                            indices=charge_region,
                            Rc=Rc,
                            mu=mu)
        w = weight.construct_weight_function()

        dv = atoms.get_volume() / calc_a.get_number_of_grid_points().prod()
        gridrefinement = 2.
        Nel = (w * (d_a + d_b)).sum() * dv / gridrefinement**3

        Zn = 0
        for atom in atoms_a:
            Zn += atom.number

        constraints.append(Zn - Nel)

    if spin_constraint:
        # can charge and spin constraints be treated at the same time?
        d_a = calc_a.get_all_electron_density(gridrefinement=2,
                                              pad=False,
                                              spin=0)
        d_b = calc_b.get_all_electron_density(gridrefinement=2,
                                              pad=False,
                                              spin=1)

        charge_region = [atom.index for atom in atoms_a]
        weight = WeightFunc(gd=gd,
                            atoms=atoms,
                            indices=charge_region,
                            Rc=Rc,
                            mu=mu)
        w = weight.construct_weight_function()

        dv = atoms.get_volume() / calc_a.get_number_of_grid_points().prod()
        gridrefinement = 2.
        Ns = (w * (d_a - d_b)).sum() * dv / gridrefinement**3

        constraints.append(Ns)

    return constraints



import numpy as np
import pytest

from ase.utils.filecache import MultiFileJSONCache
from gpaw import GPAW
from gpaw.lcao.dipoletransition import get_momentum_transitions
from gpaw.elph import ResonantRamanCalculator
from gpaw.mpi import world


def get_random_g(nk, nb):
    g_sqklnn = np.zeros((1, 1, 4, 3, 4, 4), dtype=complex)
    rng = np.random.default_rng()
    tmp = rng.random((4, 4)) + 1j * rng.random((4, 4))
    # make hermitian
    for i in range(4):
        for j in range(i + 1, 4):
            tmp[i, j] = tmp[j, i].conj()
    g_sqklnn[0, 0, 0, 2] = tmp
    return g_sqklnn


@pytest.mark.later  # calc.initialize_positions(atoms) not implemented!
@pytest.mark.serial
def test_ramancalculator(gpw_files, tmp_path_factory):
    """Test of ResonantRamanCalculator object"""
    calc = GPAW(gpw_files['bcc_li_lcao_wfs'])
    atoms = calc.atoms
    # Initialize calculator if necessary
    if not hasattr(calc.wfs, 'C_nM'):
        calc.initialize_positions(atoms)
    # need to fiddle with some occupation numnbers as this exampe is
    # not properly converged
    for kpt in calc.wfs.kpt_u:
        kpt.f_n[0] = kpt.weight

    # prepare some required data
    wph_w = np.array([0., 0., 0.1])
    get_momentum_transitions(calc.wfs)
    if world.rank == 0:
        g_sqklnn = get_random_g(4, 4)
        np.save("gsqklnn.npy", g_sqklnn)

    rrc = ResonantRamanCalculator(calc, wph_w)
    assert rrc.mom_skvnm == pytest.approx(np.transpose(rrc.mom_skvnm,
                                                       (0, 1, 2, 4, 3)).conj())
    # Force momentum matrix elements to be the same in all directions
    # else R^ab won't be R^{ba*}
    # This is a bit of a dirty hack I guess. Ideally we need a test systm with
    # equivalent axes but no degenerate bands... so yeah
    rrc.mom_skvnm[0, :, 1] = rrc.mom_skvnm[0, :, 0]
    rrc.mom_skvnm[0, :, 2] = rrc.mom_skvnm[0, :, 0]

    # check reading of file cache
    check_cache = MultiFileJSONCache("Rlab")
    assert check_cache["phonon_frequencies"] == pytest.approx(wph_w)
    assert check_cache["frequency_grid"] is None

    rrc.calculate_raman_tensor(1.0)
    for i in range(3):
        for j in range(3):
            R_l = check_cache[f"{'xyz'[i]}{'xyz'[j]}"]
            assert R_l is not None
            assert R_l[0] == pytest.approx(0.0 + 1j * 0.)
            assert R_l[1] == pytest.approx(0.0 + 1j * 0.)

        if j > i:
            # need to make sure momentum matrix is perfectly hermitian too
            Rother_l = check_cache[f"{'xyz'[j]}{'xyz'[i]}"]
            print(i, j, R_l[2], Rother_l[2])
            assert R_l[2].real == pytest.approx(Rother_l[2].real)
            assert R_l[2].imag == pytest.approx(Rother_l[2].imag)



from gpaw import GPAW


def fulldiag(filename, bands=None, scalapack=1):
    """Set up full H and S matrices and find all or some eigenvectors/values.

    calc: str
        Filename of gpw-file.
    bands: int
        Number of bands to calculate.  Defaults to all.
    scalapack: int
        Number of cores to use for ScaLapack.  Default is one.
    dry_run: bool
        Don't do actual calculation.

    """
    name, ext = filename.rsplit('.', 1)
    assert ext == 'gpw'
    calc = GPAW(filename,
                parallel={'band': scalapack},
                txt=name + '-all.txt')
    calc.diagonalize_full_hamiltonian(bands)
    calc.write(name + '-all.gpw', 'all')

    ng = calc.wfs.pd.ngmax
    mem = ng**2 * 16 / 1024**2
    print('Maximum matrix size: {0}x{0} = {1:.3f} MB'.format(ng, mem))


class CLICommand:
    """Calculate all or some eigenvectors/values for fixed H."""

    @staticmethod
    def add_arguments(parser):
        parser.add_argument('gpw', metavar='gpw-file')
        parser.add_argument('-b', '--bands', type=int)
        parser.add_argument('-s', '--scalapack', type=int, default=1)

    @staticmethod
    def run(args):
        fulldiag(args.gpw, args.bands, args.scalapack)



import pytest
from ase import Atom
from gpaw import GPAW
from gpaw.cluster import Cluster

h = 0.25
box = 3.0


@pytest.mark.later
def test_spin_spin_contamination_B():
    # B should not have spin contamination
    s = Cluster([Atom('B')])
    s.minimal_box(box, h=h)
    s.set_initial_magnetic_moments([-1])

    c = GPAW(xc='LDA', nbands=-3,
             basis='dzp',
             hund=True,
             h=h,
             # mixer=MixerDif(beta=0.05, nmaxold=5, weight=50.0),
             convergence={'eigenstates': 0.078,
                          'density': 5e-3,
                          'energy': 0.1})
    c.calculate(s)

    contamination = min(c.density.get_spin_contamination(s, 0),
                        c.density.get_spin_contamination(s, 1))

    assert contamination == pytest.approx(0.0, abs=0.01)


@pytest.mark.later
def test_spin_spin_contamination_H2():
    # setup H2 at large distance with different spins for the atoms
    s = Cluster([Atom('H'), Atom('H', [0, 0, 3.0])])
    s.minimal_box(box, h=h)
    s.set_initial_magnetic_moments([-1, 1])

    c = GPAW(xc='LDA',
             nbands=-3,
             h=h,
             convergence={'eigenstates': 0.078,
                          'density': 1e-2,
                          'energy': 0.1})
    c.calculate(s)

    scont_s = [c.density.get_spin_contamination(s),
               c.density.get_spin_contamination(s, 1)]

    assert scont_s[0] == pytest.approx(scont_s[1], abs=2e-4)  # symmetry
    assert scont_s[0] == pytest.approx(0.967, abs=2e-3)



# flake8: noqa
from gpaw.fdtd.poisson_fdtd import QSFDTD
from gpaw.fdtd.polarizable_material import PermittivityPlus, PolarizableMaterial, \
    PolarizableSphere, PolarizableBox, \
    PolarizableEllipsoid, PolarizableRod, \
    PolarizableTetrahedron
from gpaw.test import equal
import pytest


@pytest.mark.later
def test_fdtd_ed_shapes(in_tmp_dir):
    # Whole simulation cell (Angstroms)
    cell = [40, 40, 20]

    # Classical subsystem
    classical_material = PolarizableMaterial()
    classical_material.add_component(PolarizableSphere(permittivity = PermittivityPlus(data = [[1.20, 0.20, 25.0]]),
                                                       center = [10, 10, 10],
                                                       radius = 4.5))
    classical_material.add_component(PolarizableBox(permittivity = PermittivityPlus(data = [[1.40, 0.20, 25.0]]),
                                                    corner1 = [18.1, 5.1, 5.1],
                                                    corner2 = [22.9, 14.9, 14.9]))
    classical_material.add_component(PolarizableEllipsoid(permittivity = PermittivityPlus(data = [[1.60, 0.20, 25.0]]),
                                                    center  = [30.0, 10.0, 10.0],
                                                    radii   = [ 3.9, 5.9, 4.9]))
    classical_material.add_component(PolarizableRod(permittivity = PermittivityPlus(data = [[1.80, 0.20, 25.0]]),
                                                    corners = [[10.0, 21.5, 10.0], [10.0, 33.5, 10.0]],
                                                    round_corners = True,
                                                    radius = 3.9))
    classical_material.add_component(PolarizableRod(permittivity = PermittivityPlus(data = [[1.00, 0.20, 25.0]]),
                                                    corners = [[20.0, 21.5, 10.0], [25.0, 33.5, 10.0]],
                                                    round_corners = True,
                                                    radius = 2.9))
    classical_material.add_component(PolarizableTetrahedron(permittivity = PermittivityPlus(data = [[0.80, 0.20, 25.0]]),
                                                            corners = [[24.1, 16.1, 5.1],
                                                                       [30.1, 36.1, 6.1],
                                                                       [36.4, 27.6, 7.1],
                                                                       [30.0, 25.0, 14.9]]))


    # Wrap calculators
    qsfdtd = QSFDTD(classical_material = classical_material,
                    atoms              = None,
                    cells              = (cell, 2.00),
                    spacings           = [1.60, 0.40],
                    remove_moments     = (1, 1))

    # Run
    energy = qsfdtd.ground_state('gs.gpw', eigensolver = 'cg', nbands = -1)
    qsfdtd.time_propagation('gs.gpw', kick_strength=[0.000, 0.000, 0.001], time_step=10, iterations=5, dipole_moment_file='dmCl.dat')

    # Restart and run
    qsfdtd.write('td.gpw', mode='all')
    qsfdtd.time_propagation('td.gpw', kick_strength=None, time_step=10, iterations=5, dipole_moment_file='dmCl.dat')

    # Test
    ref_cl_dipole_moment = [ -1.01218549e-04,  -3.03603883e-05,   1.86063875e-01]
    tol = 1e-6
    equal(qsfdtd.td_calc.hamiltonian.poisson.get_classical_dipole_moment(), ref_cl_dipole_moment, tol)



"""
tests solvation parameters as in

O. Andreussi, I. Dabo, and N. Marzari,
The Journal of Chemical Physics, vol. 136, no. 6, p. 064102, 2012
"""

from gpaw import GPAW
from gpaw.cluster import Cluster
from gpaw.test import equal
from ase.build import molecule
from ase.units import mol, kcal, Pascal, m, Bohr
from gpaw.solvation import (
    SolvationGPAW,
    ADM12SmoothStepCavity,
    LinearDielectric,
    GradientSurface,
    SurfaceInteraction,
    KB51Volume,
    VolumeInteraction,
    ElDensity
)
from gpaw.solvation.poisson import ADM12PoissonSolver
import warnings

SKIP_VAC_CALC = True

h = 0.24
vac = 4.0

epsinf = 78.36
rhomin = 0.0001 / Bohr ** 3
rhomax = 0.0050 / Bohr ** 3
st = 50. * 1e-3 * Pascal * m
p = -0.35 * 1e9 * Pascal
convergence = {
    'energy': 0.05 / 8.,
    'density': 10.,
    'eigenstates': 10.,
}


def test_solvation_adm12():
    atoms = Cluster(molecule('H2O'))
    atoms.minimal_box(vac, h)

    if not SKIP_VAC_CALC:
        atoms.calc = GPAW(xc='PBE', h=h, convergence=convergence)
        Evac = atoms.get_potential_energy()
        print(Evac)
    else:
        # h=0.24, vac=4.0, setups: 0.9.11271, convergence: only energy 0.05 / 8
        Evac = -14.857414548

    with warnings.catch_warnings():
        # ignore production code warning for ADM12PoissonSolver
        warnings.simplefilter("ignore")
        psolver = ADM12PoissonSolver()

    atoms.calc = SolvationGPAW(
        xc='PBE', h=h, poissonsolver=psolver, convergence=convergence,
        cavity=ADM12SmoothStepCavity(
            rhomin=rhomin, rhomax=rhomax, epsinf=epsinf,
            density=ElDensity(),
            surface_calculator=GradientSurface(),
            volume_calculator=KB51Volume()
        ),
        dielectric=LinearDielectric(epsinf=epsinf),
        interactions=[
            SurfaceInteraction(surface_tension=st),
            VolumeInteraction(pressure=p)
        ]
    )
    Ewater = atoms.get_potential_energy()
    assert atoms.calc.get_number_of_iterations() < 40
    atoms.get_forces()
    DGSol = (Ewater - Evac) / (kcal / mol)
    print('Delta Gsol: %s kcal / mol' % DGSol)

    equal(DGSol, -6.3, 2.)



import pytest
import numpy as np

from ase import Atoms
from ase.parallel import parprint
from ase.units import Hartree

import gpaw.mpi as mpi
from gpaw import GPAW
from gpaw.test import equal
from gpaw.lrtddft.kssingle import KSSingles


@pytest.mark.lrtddft
def test_lrtddft_kssingles_Be(in_tmp_dir):
    Be = Atoms('Be')
    Be.center(vacuum=4)
    if 1:
        # introduce a sligth non-orthgonality
        cell = Be.get_cell()
        cell[1] += 0.001 * cell[0]
        Be.set_cell(cell)

    txt = None
    eigensolver = None

    # modes = ['lcao', 'fd']
    modes = ['fd']

    for mode in modes:
        energy = {}
        osz = {}
        for pbc in [False, True]:
            Be.set_pbc(pbc)
            if pbc:
                name = 'periodic'
                calc = GPAW(h=0.25,
                            nbands=4,
                            kpts=(1, 2, 2),
                            mode=mode,
                            poissonsolver={'name': 'fd'},
                            symmetry='off',
                            eigensolver=eigensolver,
                            txt=txt)
            else:
                name = 'zero_bc'
                calc = GPAW(h=0.25, nbands=4, mode=mode,
                            poissonsolver={'name': 'fd'},
                            eigensolver=eigensolver, txt=txt)
            Be.calc = calc
            Be.get_potential_energy()

            kss = KSSingles(restrict={'eps': 0.9})
            kss.calculate(Be)
            # all s->p transitions at the same energy [Ha] and
            # oscillator_strength
            for ks in kss:
                equal(ks.get_energy(), kss[0].get_energy(), 5.e-3)
                equal(ks.get_oscillator_strength()[0],
                      kss[0].get_oscillator_strength()[0], 5.e-3)
                equal(ks.get_oscillator_strength()[0],
                      ks.get_oscillator_strength()[1:].sum() / 3, 1.e-15)
                for c in range(3):
                    equal(ks.get_oscillator_strength()[1 + c],
                          ks.get_dipole_tensor()[c, c], 1.e-15)
            energy[name] = np.array(
                [ks.get_energy() * Hartree for ks in kss]).mean()
            osz[name] = np.array(
                [ks.get_oscillator_strength()[0] for ks in kss]).sum()

            parprint(name + ':')
            parprint(kss)

            # I/O
            fname = 'kss_' + name + '.dat'
            kss.write(fname)
            mpi.world.barrier()
            kss = KSSingles.read(fname)
            kss1 = KSSingles.read(fname, restrict={'jend': 1})
            assert len(kss1) == calc.wfs.kd.nibzkpts * calc.wfs.nspins

        # periodic and non-periodic should be roughly equal
        equal(energy['zero_bc'], energy['periodic'], 5.e-2)
        equal(osz['zero_bc'], osz['periodic'], 2.e-2)



import numpy as np
from ase import Atoms
from gpaw import GPAW
from gpaw.test import equal


def test_lcao_bulk(in_tmp_dir):
    bulk = Atoms('Li', pbc=True)
    k = 4
    g = 8
    calc = GPAW(gpts=(g, g, g), kpts=(k, k, k),
                mode='lcao', basis='dzp')
    bulk.calc = calc
    e = []
    niter = []
    A = [2.6, 2.65, 2.7, 2.75, 2.8]
    for a in A:
        bulk.set_cell((a, a, a))
        e.append(bulk.get_potential_energy())
        niter.append(calc.get_number_of_iterations())

    a = np.roots(np.polyder(np.polyfit(A, e, 2), 1))[0]
    print('a =', a)
    equal(a, 2.6378, 0.0003)

    e_ref = [-1.8677343236247692, -1.8690343169380492, -1.8654175796625045,
             -1.8566274574918875, -1.8432374955346396]
    niter_ref = [6, 6, 6, 6, 6]

    print(e)
    energy_tolerance = 0.0003
    niter_tolerance = 0

    for i in range(len(A)):
        equal(e[i], e_ref[i], energy_tolerance)
        equal(niter[i], niter_ref[i], niter_tolerance)

    wf1 = calc.get_pseudo_wave_function(kpt=3, band=0)
    calc.write('Li', mode='all')
    calc2 = GPAW('Li')
    # calc2.initialize_positions()
    wf2 = calc2.get_pseudo_wave_function(kpt=3, band=0)
    equal(abs(wf1 - wf2).max(), 0, 1e-9)



import warnings

import pytest
from ase.build import molecule

from gpaw import GPAW

# Silence those warnings.


@pytest.mark.legacy
def test_fileio_idiotproof_setup(in_tmp_dir):
    warnings.filterwarnings('ignore', 'Setup for',)

    m = molecule('H')
    m.center(vacuum=2.0)
    calc = GPAW(mode='lcao')
    m.calc = calc
    m.get_potential_energy()
    calc.write('r.gpw')
    calc = GPAW('r.gpw', xc='PBE')
    calc.get_potential_energy()



import time

import ase.units as units
import numpy as np
import pytest
from ase import Atoms
from ase.calculators.tip3p import TIP3P, angleHOH, rOH
from ase.constraints import FixBondLengths
from ase.io import read
from ase.io.trajectory import Trajectory
from ase.md import Langevin as Langevin0

from gpaw.utilities.watermodel import FixBondLengthsWaterModel, TIP3PWaterModel


def Langevin(*args, **kwargs):
    try:
        return Langevin0(*args, **kwargs)
    except TypeError:
        kT = kwargs.pop('temperature_K') * units.kB
        return Langevin0(*args, **kwargs, temperature=kT)


@pytest.mark.slow
def test_watermodel(in_tmp_dir):
    NSTEPS = 600
    SCALE = 200

    cutoff = 4.0

    # Set up water box at 20 deg C density
    x = angleHOH * np.pi / 180 / 2
    pos = [[0, 0, 0],
           [0, rOH * np.cos(x), rOH * np.sin(x)],
           [0, rOH * np.cos(x), -rOH * np.sin(x)]]
    atoms = Atoms('OH2', positions=pos)

    vol = ((18.01528 / 6.022140857e23) / (0.9982 / 1e24))**(1 / 3.)
    atoms.set_cell((vol, vol, vol))
    atoms.center()

    N = 4
    atoms = atoms.repeat((N, N, N))
    atoms.set_pbc(True)

    pairs = [(3 * i + j, 3 * i + (j + 1) % 3)
             for i in range(len(atoms) // 3)
             for j in [0, 1, 2]]

    # Create atoms object with old constraints for reference
    atoms_ref = atoms.copy()
    atoms_ref.constraints = FixBondLengths(pairs)

    # RATTLE-type constraints on O-H1, O-H2, H1-H2.
    atoms.constraints = FixBondLengthsWaterModel(pairs)

    atoms.calc = TIP3PWaterModel(rc=cutoff)
    atoms_ref.calc = TIP3P(rc=cutoff)

    np.random.seed(123)
    md = Langevin(atoms, 1 * units.fs, temperature_K=300,
                  friction=0.01, logfile='C.log')
    traj = Trajectory('C.traj', 'w', atoms)
    md.attach(traj.write, interval=1)

    start = time.time()
    with md:
        md.run(NSTEPS)
    end = time.time()
    Cversion = end - start
    print("%d steps of C-MD took %.3fs (%.0f ms/step)" % (
        NSTEPS, Cversion,
        Cversion / NSTEPS * 1000))
    traj.close()

    np.random.seed(123)
    md_ref = Langevin(atoms_ref, 1 * units.fs, temperature_K=300,
                      friction=0.01, logfile='ref.log')
    traj_ref = Trajectory('ref.traj', 'w', atoms_ref)
    md_ref.attach(traj_ref.write, interval=1)
    start = time.time()
    with md_ref:
        md_ref.run(NSTEPS / SCALE)
    end = time.time()
    Pyversion = (end - start) * SCALE
    print("%d steps of Py-MD took %.3fs (%.0f ms/step)" % (
        NSTEPS / SCALE,
        Pyversion / SCALE, Pyversion / NSTEPS * 1000))
    traj_ref.close()

    # Compare trajectories
    images = read('C.traj@:')
    images_ref = read('ref.traj@:')
    for img1, img2 in zip(images, images_ref):
        norm = np.linalg.norm(img1.get_positions() - img2.get_positions())
        print(norm)
        assert norm < 1e-11

    print("Speedup", Pyversion / Cversion)



"""This module wraps the gpaw preconditioner for use with scipy solvers."""

import numpy as np

from gpaw.fd_operators import Laplace
from gpaw.preconditioner import Preconditioner


class ScipyPreconditioner:

    def __init__(self, gd, project, dtype=float):
        """Init the gpaw preconditioner.

        Parameters
        ----------
        gd: GridDescriptor
            Coarse grid

        """

        self.project = project
        self.gd = gd
        # K-point for the preconditioner
        self.kpt = None

        kin = Laplace(gd, scale=-0.5, n=3, dtype=dtype)
        self.pc = Preconditioner(gd, kin, dtype=dtype)

        # For scipy's linear solver
        N = np.prod(gd.n_c)
        self.shape = (N, N)
        self.dtype = dtype

    def set_kpt(self, kpt):
        """Set k-point for ``Transformer`` objects inside the preconditioner.

        Parameters
        ----------
        kpt: KPoint or KPointContainer
            Only requirement is that it must have an ``phase_cd`` attribute.

        """

        self.kpt = kpt

    def matvec(self, x):
        """Matrix vector multiplication for ``scipy.sparse.linalg`` solvers.

        Parameters
        ----------
        x: ndarry
            1-dimensional array holding the grid representation of the vector

        """

        # Output array
        y_G = self.gd.zeros(dtype=self.dtype)
        shape = y_G.shape

        size = x.size
        assert size == np.prod(shape)

        x_G = x.reshape(shape)

        # Call gpaw preconditioner
        y_G = self.pc(x_G, kpt=self.kpt)

        # Project out undesired (numerical) components
        self.project(y_G)

        y = y_G.ravel()

        return y



import pytest
from itertools import product

import numpy as np

from gpaw import GPAW
from gpaw.mpi import world
from gpaw.response import ResponseContext, ResponseGroundStateAdapter
from gpaw.response.pw_parallelization import block_partition
from gpaw.response.kspair import KohnShamKPointPairExtractor
from gpaw.response.pair_functions import SingleQPWDescriptor
from gpaw.response.pair_transitions import PairTransitions
from gpaw.response.pair_integrator import KPointPairPointIntegral
from gpaw.response.symmetry import PWSymmetryAnalyzer

from gpaw.test.response.test_chiks import (generate_system_s, generate_qrel_q,
                                           get_q_c, generate_nblocks_n)
from gpaw.test.conftest import response_band_cutoff

pytestmark = pytest.mark.skipif(world.size == 1, reason='world.size == 1')


# ---------- Actual tests ---------- #


@pytest.mark.response
@pytest.mark.parametrize('system,qrel,nblocks', product(generate_system_s(),
                                                        generate_qrel_q(),
                                                        generate_nblocks_n()))
def test_parallel_extract_kptdata(in_tmp_dir, gpw_files,
                                  system, qrel, nblocks):
    """Test that the KohnShamKPointPair data extracted from a serial and a
    parallel calculator object is identical."""

    # ---------- Inputs ---------- #

    wfs, spincomponent = system
    q_c = get_q_c(wfs, qrel)

    # ---------- Script ---------- #

    # Initialize serial ground state adapter
    context = ResponseContext()
    serial_gs = ResponseGroundStateAdapter.from_gpw_file(
        gpw_files[wfs], context)
    
    # Initialize parallel ground state adapter
    calc = GPAW(gpw_files[wfs], parallel=dict(domain=1))
    nbands = response_band_cutoff[wfs]
    parallel_gs = ResponseGroundStateAdapter(calc)

    # Set up extractors and integrals
    tcomm, kcomm = block_partition(context.comm, nblocks)
    serial_extractor = initialize_extractor(
        serial_gs, context, tcomm, kcomm)
    parallel_extractor = initialize_extractor(
        parallel_gs, context, tcomm, kcomm)
    serial_integral = initialize_integral(serial_extractor, context, q_c)
    parallel_integral = initialize_integral(parallel_extractor, context, q_c)

    # Set up transitions
    transitions = initialize_transitions(
        serial_extractor, spincomponent, nbands)

    # Extract and compare kptpairs
    ni = serial_integral.ni  # Number of iterations in kptpair generator
    assert parallel_integral.ni == ni
    serial_kptpairs = serial_integral.weighted_kpoint_pairs(transitions)
    parallel_kptpairs = parallel_integral.weighted_kpoint_pairs(transitions)
    for _ in range(ni):
        kptpair1, _ = next(serial_kptpairs)
        kptpair2, _ = next(parallel_kptpairs)
        compare_kptpairs(kptpair1, kptpair2)


# ---------- Test functionality ---------- #


def compare_kptpairs(kptpair1, kptpair2):
    if kptpair1 is None:
        # Due to k-point distribution, all ranks don't necessarily have a
        # kptpair to integrate
        assert kptpair2 is None
        return
    assert kptpair1.K1 == kptpair2.K1
    assert kptpair1.K2 == kptpair2.K2
    assert np.allclose(kptpair1.deps_t, kptpair2.deps_t)
    assert np.allclose(kptpair1.df_t, kptpair2.df_t)

    compare_ikpts(kptpair1.ikpt1, kptpair2.ikpt1)
    compare_ikpts(kptpair1.ikpt2, kptpair2.ikpt2)


def compare_ikpts(ikpt1, ikpt2):
    assert ikpt1.ik == ikpt2.ik
    assert np.allclose(ikpt1.Ph.array, ikpt2.Ph.array)
    assert np.allclose(ikpt1.psit_hG, ikpt2.psit_hG)
    assert np.all(ikpt1.h_myt == ikpt2.h_myt)


def initialize_extractor(gs, context, tcomm, kcomm):
    return KohnShamKPointPairExtractor(gs, context,
                                       transitions_blockcomm=tcomm,
                                       kpts_blockcomm=kcomm)


def initialize_integral(extractor, context, q_c):
    # Initialize symmetry analyzer
    gs = extractor.gs
    qpd = SingleQPWDescriptor.from_q(q_c, 1e-3, gs.gd)
    analyzer = PWSymmetryAnalyzer(gs.kd, qpd, context)

    return KPointPairPointIntegral(extractor, analyzer)


def initialize_transitions(extractor, spincomponent, nbands):
    bandsummation = 'pairwise'
    return PairTransitions.from_transitions_domain_arguments(
        spincomponent, nbands, extractor.nocc1, extractor.nocc2,
        extractor.gs.nspins, bandsummation)



from __future__ import annotations
import contextlib
from time import time
from typing import TYPE_CHECKING

import numpy as np

cupy_is_fake = True
"""True if :mod:`cupy` has been replaced by ``gpaw.gpu.cpupy``"""

is_hip = False
"""True if we are using HIP"""

if TYPE_CHECKING:
    import gpaw.gpu.cpupy as cupy
    import gpaw.gpu.cpupyx as cupyx
else:
    try:
        import cupy
        import cupyx
        from cupy.cuda import runtime
        is_hip = runtime.is_hip
        cupy_is_fake = False
    except ImportError:
        import gpaw.gpu.cpupy as cupy
        import gpaw.gpu.cpupyx as cupyx

__all__ = ['cupy', 'cupyx', 'as_xp', 'synchronize']


def synchronize():
    if not cupy_is_fake:
        cupy.cuda.runtime.deviceSynchronize()


def setup():
    # select GPU device (round-robin based on MPI rank)
    # if not set, all MPI ranks will use the same default device
    if not cupy_is_fake:
        from gpaw.mpi import rank
        device_id = rank % cupy.cuda.runtime.getDeviceCount()
        cupy.cuda.runtime.setDevice(device_id)


def as_xp(array, xp):
    """Transfer array to CPU or GPU (if not already there).

    Parameters
    ==========
    array:
        Numpy or CuPy array.
    xp:
        :mod:`numpy` or :mod:`cupy`.
    """
    if xp is np:
        if isinstance(array, np.ndarray):
            return array
        return cupy.asnumpy(array)
    if isinstance(array, np.ndarray):
        return cupy.asarray(array)
    1 / 0
    return array


def cupy_eigh(a: cupy.ndarray, UPLO: str) -> tuple[cupy.ndarray, cupy.ndarray]:
    """Wrapper for ``eigh()``.

    HIP-GPU version is too slow for now so we do it on the CPU.
    """
    from scipy.linalg import eigh
    if not is_hip:
        return cupy.linalg.eigh(a, UPLO=UPLO)
    eigs, evals = eigh(cupy.asnumpy(a), lower=(UPLO == 'L'))
    return cupy.asarray(eigs), cupy.asarray(evals)


@contextlib.contextmanager
def T():
    t1 = time()
    yield
    synchronize()
    t2 = time()
    print(f'{(t2 - t1) * 1e9:_.3f} ns')



"""Analytic expressions of the homogeneous electron gas (HEG).

The goal of this module is to centralize all the standard expressions
needed around GPAW for doing HEG things.  Please move appropriate
code here whenever relevant."""

import numpy as np
from ase.utils import lazyproperty


class HEG:
    def __init__(self, rs):
        self._rs = rs

    @property
    def rs(self):
        return self._rs

    @lazyproperty
    def qF(self):
        return (9.0 * np.pi / 4.0)**(1.0 / 3.0) / self.rs



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""This module provides all the classes and functions associated with the
evaluation of exact exchange.
"""

from math import exp, ceil
import numpy as np

from gpaw.atom.configurations import core_states
from gpaw.gaunt import gaunt
from gpaw.lfc import LFC
from gpaw.poisson import PoissonSolver
from gpaw.helmholtz import HelmholtzSolver
from gpaw.transformers import Transformer
from gpaw.utilities import hartree, pack, pack2, unpack, unpack2, packed_index
from gpaw.utilities.blas import mmm
from gpaw.utilities.tools import symmetrize
from gpaw.xc import XC
from gpaw.xc.functional import XCFunctional
from gpaw.xc.kernel import XCNull


class HybridXCBase(XCFunctional):
    orbital_dependent = True
    omega = None

    def __init__(self, name, stencil=2, hybrid=None, xc=None, omega=None):
        """Mix standard functionals with exact exchange.

        name: str
            Name of hybrid functional.
        hybrid: float
            Fraction of exact exchange.
        xc: str or XCFunctional object
            Standard DFT functional with scaled down exchange.
        """

        rsf_functionals = {    # Parameters can also be taken from libxc
            'CAMY-BLYP': {  # Akinaga, Ten-no CPL 462 (2008) 348-351
                'alpha': 0.2,
                'beta': 0.8,
                'omega': 0.44,
                'cam': True,
                'rsf': 'Yukawa',
                'xc': 'HYB_GGA_XC_CAMY_BLYP'
            },
            'CAMY-B3LYP': {  # Seth, Ziegler JCTC 8 (2012) 901-907
                'alpha': 0.19,
                'beta': 0.46,
                'omega': 0.34,
                'cam': True,
                'rsf': 'Yukawa',
                'xc': 'HYB_GGA_XC_CAMY_B3LYP'
            },
            'LCY-BLYP': {  # Seth, Ziegler JCTC 8 (2012) 901-907
                'alpha': 0.0,
                'beta': 1.0,
                'omega': 0.75,
                'cam': False,
                'rsf': 'Yukawa',
                'xc': 'HYB_GGA_XC_LCY_BLYP'
            },
            'LCY-PBE': {  # Seth, Ziegler JCTC 8 (2012) 901-907
                'alpha': 0.0,
                'beta': 1.0,
                'omega': 0.75,
                'cam': False,
                'rsf': 'Yukawa',
                'xc': 'HYB_GGA_XC_LCY_PBE'
            }
        }
        self.omega = None
        self.cam_alpha = None
        self.cam_beta = None
        self.is_cam = False
        self.rsf = None

        def _xc(name):
            return {'name': name, 'stencil': stencil}

        if name == 'EXX':
            hybrid = 1.0
            xc = XC(XCNull())
        elif name == 'PBE0':
            hybrid = 0.25
            xc = XC(_xc('HYB_GGA_XC_PBEH'))
        elif name == 'B3LYP':
            hybrid = 0.2
            xc = XC(_xc('HYB_GGA_XC_B3LYP'))
        elif name in rsf_functionals:
            rsf_functional = rsf_functionals[name]
            self.cam_alpha = rsf_functional['alpha']
            self.cam_beta = rsf_functional['beta']
            self.omega = rsf_functional['omega']
            self.is_cam = rsf_functional['cam']
            self.rsf = rsf_functional['rsf']
            xc = XC(rsf_functional['xc'])
            hybrid = self.cam_alpha + self.cam_beta

        if isinstance(xc, (str, dict)):
            xc = XC(xc)

        self.hybrid = float(hybrid)
        self.xc = xc
        if omega is not None:
            omega = float(omega)
            if self.omega is not None and self.omega != omega:
                self.xc.kernel.set_omega(omega)
                # Needed to tune omega for RSF
            self.omega = omega
        XCFunctional.__init__(self, name, xc.type)

    def todict(self):
        return {'name': self.name,
                'hybrid': self.hybrid,
                'excitation': self.excitation,
                'excited': self.excited,
                'xc': self.xc.todict(),
                'omega': self.omega}

    def tostring(self):
        """Return string suitable to generate xc from string."""
        xc_dict = self.todict()
        for test_key in ['name', 'xc', 'kernel', 'type']:
            if test_key in xc_dict:
                del xc_dict[test_key]
        return self.name + ':' + ':'.join([(k + '=' + repr(v))
                                           for k, v in xc_dict.items()])

    def get_setup_name(self):
        return 'PBE'


class HybridXC(HybridXCBase):
    def __init__(self, name, hybrid=None, xc=None,
                 finegrid=False, unocc=False, omega=None,
                 excitation=None, excited=0, stencil=2):
        """Mix standard functionals with exact exchange.

        finegrid: boolean
            Use fine grid for energy functional evaluations ?
        unocc: boolean
            Apply vxx also to unoccupied states ?
        omega: float
            RSF mixing parameter
        excitation: string:
            Apply operator for improved virtual orbitals
            to unocc states? Possible modes:
                singlet: excitations to singlets
                triplet: excitations to triplets
                average: average between singlets and tripletts
                see f.e. https://doi.org/10.1021/acs.jctc.8b00238
        excited: number
            Band to excite from - counted from HOMO downwards

        """
        self.finegrid = finegrid
        self.unocc = unocc
        self.excitation = excitation
        self.excited = excited
        HybridXCBase.__init__(self, name, hybrid=hybrid, xc=xc, omega=omega,
                              stencil=stencil)

    def calculate_paw_correction(self, setup, D_sp, dEdD_sp=None,
                                 addcoredensity=True, a=None):
        return self.xc.calculate_paw_correction(setup, D_sp, dEdD_sp,
                                                addcoredensity, a)

    def initialize(self, density, hamiltonian, wfs):
        assert wfs.kd.gamma
        self.xc.initialize(density, hamiltonian, wfs)
        self.kpt_comm = wfs.kd.comm
        self.nspins = wfs.nspins
        self.setups = wfs.setups
        self.density = density
        self.kpt_u = wfs.kpt_u
        self.exx_s = np.zeros(self.nspins)
        self.ekin_s = np.zeros(self.nspins)
        self.nocc_s = np.empty(self.nspins, int)

        self.gd = density.gd
        self.redistributor = density.redistributor

        use_charge_center = hamiltonian.poisson.use_charge_center
        # XXX How do we construct a copy of the Poisson solver of the
        # Hamiltonian?  We don't know what class it is, etc., but gd
        # may differ.
        # XXX One might consider using a charged centered compensation
        # charge for the PoissonSolver in the case of EXX as standard
        self.poissonsolver = PoissonSolver(
            'fd', eps=1e-12, use_charge_center=use_charge_center)
        # self.poissonsolver = hamiltonian.poisson

        if self.finegrid:
            self.finegd = self.gd.refine()
            # XXX Taking restrictor from Hamiltonian will not work in PW mode,
            # will it?  I think this supports only real-space mode.
            # self.restrictor = hamiltonian.restrictor
            self.restrictor = Transformer(self.finegd, self.gd, 3)
            self.interpolator = Transformer(self.gd, self.finegd, 3)
        else:
            self.finegd = self.gd

        self.ghat = LFC(self.finegd,
                        [setup.ghat_l for setup in density.setups],
                        integral=np.sqrt(4 * np.pi), forces=True)
        self.poissonsolver.set_grid_descriptor(self.finegd)
        if self.rsf == 'Yukawa':
            omega2 = self.omega**2
            self.screened_poissonsolver = HelmholtzSolver(
                k2=-omega2, eps=1e-12, nn=3,
                use_charge_center=use_charge_center)
            self.screened_poissonsolver.set_grid_descriptor(self.finegd)

    def set_positions(self, spos_ac):
        self.ghat.set_positions(spos_ac)

    def calculate(self, gd, n_sg, v_sg=None, e_g=None):
        # Normal XC contribution:
        exc = self.xc.calculate(gd, n_sg, v_sg, e_g)
        # Note that the quantities passed are on the
        # density/Hamiltonian grids!
        # They may be distributed differently from own quantities.
        self.ekin = self.kpt_comm.sum(self.ekin_s.sum())
        return exc + self.kpt_comm.sum(self.exx_s.sum())

    def calculate_exx(self):
        for kpt in self.kpt_u:
            self.apply_orbital_dependent_hamiltonian(kpt, kpt.psit_nG)

    def apply_orbital_dependent_hamiltonian(self, kpt, psit_nG,
                                            Htpsit_nG=None, dH_asp=None):
        if kpt.f_n is None:
            return

        deg = 2 // self.nspins   # Spin degeneracy
        hybrid = self.hybrid
        P_ani = kpt.P_ani
        setups = self.setups
        is_cam = self.is_cam

        vt_g = self.finegd.empty()
        if self.gd is not self.finegd:
            vt_G = self.gd.empty()
        if self.rsf == 'Yukawa':
            y_vt_g = self.finegd.empty()
            # if self.gd is not self.finegd:
            #     y_vt_G = self.gd.empty()

        nocc = int(ceil(kpt.f_n.sum())) // (3 - self.nspins)
        if self.excitation is not None:
            ex_band = nocc - self.excited - 1
            if self.excitation == 'singlet':
                ex_weight = -1
            elif self.excitation == 'triplet':
                ex_weight = +1
            else:
                ex_weight = 0

        if self.unocc or self.excitation is not None:
            nbands = len(kpt.f_n)
        else:
            nbands = nocc
        self.nocc_s[kpt.s] = nocc

        if Htpsit_nG is not None:
            kpt.vt_nG = self.gd.empty(nbands)
            kpt.vxx_ani = {}
            kpt.vxx_anii = {}
            for a, P_ni in P_ani.items():
                I = P_ni.shape[1]
                kpt.vxx_ani[a] = np.zeros((nbands, I))
                kpt.vxx_anii[a] = np.zeros((nbands, I, I))

        exx = 0.0
        ekin = 0.0

        # XXXX nbands can be different numbers on different cpus!
        # That means some will execute the loop and others not.
        # And deadlocks with augment-grids.

        # Determine pseudo-exchange
        for n1 in range(nbands):
            psit1_G = psit_nG[n1]
            f1 = kpt.f_n[n1] / deg
            for n2 in range(n1, nbands):
                psit2_G = psit_nG[n2]
                f2 = kpt.f_n[n2] / deg
                if n1 != n2 and f1 == 0 and f1 == f2:
                    continue    # Don't work on double unocc. bands
                # Double count factor:
                dc = (1 + (n1 != n2)) * deg
                nt_G, rhot_g = self.calculate_pair_density(n1, n2, psit_nG,
                                                           P_ani)
                vt_g[:] = 0.0
                # XXXXX This will go wrong because we are solving the
                # Poisson equation on the distribution of gd, not finegd
                # Or maybe it's fixed now

                self.poissonsolver.solve(vt_g, -rhot_g,
                                         charge=-float(n1 == n2),
                                         zero_initial_phi=True)
                vt_g *= hybrid
                if self.rsf == 'Yukawa':
                    y_vt_g[:] = 0.0
                    self.screened_poissonsolver.solve(
                        y_vt_g, -rhot_g, charge=-float(n1 == n2),
                        zero_initial_phi=True)
                    if is_cam:  # Cam like correction
                        y_vt_g *= self.cam_beta
                    else:
                        y_vt_g *= hybrid
                    vt_g -= y_vt_g
                if self.gd is self.finegd:
                    vt_G = vt_g
                else:
                    self.restrictor.apply(vt_g, vt_G)

                # Integrate the potential on fine and coarse grids
                int_fine = self.finegd.integrate(vt_g * rhot_g)
                int_coarse = self.gd.integrate(vt_G * nt_G)
                if self.gd.comm.rank == 0:  # only add to energy on master CPU
                    exx += 0.5 * dc * f1 * f2 * int_fine
                    ekin -= dc * f1 * f2 * int_coarse
                if Htpsit_nG is not None:
                    Htpsit_nG[n1] += f2 * vt_G * psit2_G
                    if n1 == n2:
                        kpt.vt_nG[n1] = f1 * vt_G
                        if self.excitation is not None and n1 == ex_band:
                            Htpsit_nG[nocc:] += f1 * vt_G * psit_nG[nocc:]
                    else:
                        if self.excitation is None or n1 != ex_band \
                                or n2 < nocc:
                            Htpsit_nG[n2] += f1 * vt_G * psit1_G
                        else:
                            Htpsit_nG[n2] += f1 * ex_weight * vt_G * psit1_G

                    # Update the vxx_uni and vxx_unii vectors of the nuclei,
                    # used to determine the atomic hamiltonian, and the
                    # residuals
                    v_aL = self.ghat.dict()
                    self.ghat.integrate(vt_g, v_aL)
                    for a, v_L in v_aL.items():
                        v_ii = unpack(np.dot(setups[a].Delta_pL, v_L))
                        v_ni = kpt.vxx_ani[a]
                        v_nii = kpt.vxx_anii[a]
                        P_ni = P_ani[a]
                        v_ni[n1] += f2 * np.dot(v_ii, P_ni[n2])
                        if n1 != n2:
                            if self.excitation is None or n1 != ex_band or \
                                    n2 < nocc:
                                v_ni[n2] += f1 * np.dot(v_ii, P_ni[n1])
                            else:
                                v_ni[n2] += f1 * ex_weight * \
                                    np.dot(v_ii, P_ni[n1])
                        else:
                            # XXX Check this:
                            v_nii[n1] = f1 * v_ii
                            if self.excitation is not None and n1 == ex_band:
                                for nuoc in range(nocc, nbands):
                                    v_ni[nuoc] += f1 * \
                                        np.dot(v_ii, P_ni[nuoc])

        def calculate_vv(ni, D_ii, M_pp, weight, addme=False):
            """Calculate the local corrections depending on Mpp."""
            dexx = 0
            dekin = 0
            if not addme:
                addsign = -2.0
            else:
                addsign = 2.0
            for i1 in range(ni):
                for i2 in range(ni):
                    A = 0.0
                    for i3 in range(ni):
                        p13 = packed_index(i1, i3, ni)
                        for i4 in range(ni):
                            p24 = packed_index(i2, i4, ni)
                            A += M_pp[p13, p24] * D_ii[i3, i4]
                    p12 = packed_index(i1, i2, ni)
                    if Htpsit_nG is not None:
                        dH_p[p12] += addsign * weight / \
                            deg * A / ((i1 != i2) + 1)
                    dekin += 2 * weight / deg * D_ii[i1, i2] * A
                    dexx -= weight / deg * D_ii[i1, i2] * A
            return (dexx, dekin)

        # Apply the atomic corrections to the energy and the Hamiltonian
        # matrix
        for a, P_ni in P_ani.items():
            setup = setups[a]

            if Htpsit_nG is not None:
                # Add non-trivial corrections the Hamiltonian matrix
                h_nn = symmetrize(np.inner(P_ni[:nbands],
                                           kpt.vxx_ani[a][:nbands]))
                ekin -= np.dot(kpt.f_n[:nbands], h_nn.diagonal())

                dH_p = dH_asp[a][kpt.s]

            # Get atomic density and Hamiltonian matrices
            D_p = self.density.D_asp[a][kpt.s]
            D_ii = unpack2(D_p)
            ni = len(D_ii)

            # Add atomic corrections to the valence-valence exchange energy
            # --
            # >  D   C     D
            # --  ii  iiii  ii
            (dexx, dekin) = calculate_vv(ni, D_ii, setup.M_pp, hybrid)
            ekin += dekin
            exx += dexx
            if self.rsf is not None:
                Mg_pp = setup.calculate_yukawa_interaction(self.omega)
                if is_cam:
                    (dexx, dekin) = calculate_vv(
                        ni, D_ii, Mg_pp, self.cam_beta, addme=True)
                else:
                    (dexx, dekin) = calculate_vv(
                        ni, D_ii, Mg_pp, hybrid, addme=True)
                ekin -= dekin
                exx -= dexx
            # Add valence-core exchange energy
            # --
            # >  X   D
            # --  ii  ii
            if setup.X_p is not None:
                exx -= hybrid * np.dot(D_p, setup.X_p)
                if Htpsit_nG is not None:
                    dH_p -= hybrid * setup.X_p
                    ekin += hybrid * np.dot(D_p, setup.X_p)

                if self.rsf == 'Yukawa' and setup.X_pg is not None:
                    if is_cam:
                        thybrid = self.cam_beta  # 0th order
                    else:
                        thybrid = hybrid
                    exx += thybrid * np.dot(D_p, setup.X_pg)
                    if Htpsit_nG is not None:
                        dH_p += thybrid * setup.X_pg
                        ekin -= thybrid * np.dot(D_p, setup.X_pg)
                elif self.rsf == 'Yukawa' and setup.X_pg is None:
                    thybrid = exp(-3.62e-2 * self.omega)  # educated guess
                    if is_cam:
                        thybrid *= self.cam_beta
                    else:
                        thybrid *= hybrid
                    exx += thybrid * np.dot(D_p, setup.X_p)
                    if Htpsit_nG is not None:
                        dH_p += thybrid * setup.X_p
                        ekin -= thybrid * np.dot(D_p, setup.X_p)
                # Add core-core exchange energy
                if kpt.s == 0:
                    if self.rsf is None or is_cam:
                        if is_cam:
                            exx += self.cam_alpha * setup.ExxC
                        else:
                            exx += hybrid * setup.ExxC

        self.exx_s[kpt.s] = self.gd.comm.sum(exx)
        self.ekin_s[kpt.s] = self.gd.comm.sum(ekin)

    def correct_hamiltonian_matrix(self, kpt, H_nn):
        if not hasattr(kpt, 'vxx_ani'):
            return

        # if self.gd.comm.rank > 0:
        #    H_nn[:] = 0.0

        nocc = self.nocc_s[kpt.s]
        nbands = len(kpt.vt_nG)
        for a, P_ni in kpt.P_ani.items():
            H_nn[:nbands, :nbands] += symmetrize(np.inner(P_ni[:nbands],
                                                          kpt.vxx_ani[a]))
        # self.gd.comm.sum(H_nn)

        if not self.unocc or self.excitation is not None:
            H_nn[:nocc, nocc:] = 0.0
            H_nn[nocc:, :nocc] = 0.0

    def calculate_pair_density(self, n1, n2, psit_nG, P_ani):
        Q_aL = {}
        for a, P_ni in P_ani.items():
            P1_i = P_ni[n1]
            P2_i = P_ni[n2]
            D_ii = np.outer(P1_i, P2_i.conj()).real
            D_p = pack(D_ii)
            Q_aL[a] = np.dot(D_p, self.setups[a].Delta_pL)

        nt_G = psit_nG[n1] * psit_nG[n2]

        if self.finegd is self.gd:
            nt_g = nt_G
        else:
            nt_g = self.finegd.empty()
            self.interpolator.apply(nt_G, nt_g)

        rhot_g = nt_g.copy()
        self.ghat.add(rhot_g, Q_aL)

        return nt_G, rhot_g

    def add_correction(self, kpt, psit_xG, Htpsit_xG, P_axi, c_axi, n_x,
                       calculate_change=False):
        if kpt.f_n is None:
            return

        if self.unocc or self.excitation is not None:
            nocc = len(kpt.vt_nG)
        else:
            nocc = self.nocc_s[kpt.s]

        if calculate_change:
            for x, n in enumerate(n_x):
                if n < nocc:
                    Htpsit_xG[x] += kpt.vt_nG[n] * psit_xG[x]
                    for a, P_xi in P_axi.items():
                        c_axi[a][x] += np.dot(kpt.vxx_anii[a][n], P_xi[x])
        else:
            for a, c_xi in c_axi.items():
                c_xi[:nocc] += kpt.vxx_ani[a][:nocc]

    def rotate(self, kpt, U_nn):
        if kpt.f_n is None:
            return

        U_nn = U_nn.T.copy()
        nocc = self.nocc_s[kpt.s]
        if len(kpt.vt_nG) == nocc:
            U_nn = U_nn[:nocc, :nocc]
        # gemm(1.0, kpt.vt_nG.copy(), U_nn, 0.0, kpt.vt_nG)
        n, G1, G2, G3 = kpt.vt_nG.shape
        vt_nG = kpt.vt_nG.reshape((n, G1 * G2 * G3))
        mmm(1.0, U_nn, 'N', vt_nG.copy(), 'N', 0.0, vt_nG)
        for v_ni in kpt.vxx_ani.values():
            # gemm(1.0, v_ni.copy(), U_nn, 0.0, v_ni)
            mmm(1.0, U_nn, 'N', v_ni.copy(), 'N', 0.0, v_ni)
        for v_nii in kpt.vxx_anii.values():
            # gemm(1.0, v_nii.copy(), U_nn, 0.0, v_nii)
            n, i, i = v_nii.shape
            v_nii = v_nii.reshape((n, i**2))
            mmm(1.0, U_nn, 'N', v_nii.copy(), 'N', 0.0, v_nii)


def atomic_exact_exchange(atom, type='all'):
    """Returns the exact exchange energy of the atom defined by the
       instantiated AllElectron object 'atom'
    """
    G_LLL = gaunt(lmax=max(atom.l_j))  # Make gaunt coeff. list
    Nj = len(atom.n_j)  # The total number of orbitals

    # determine relevant states for chosen type of exchange contribution
    if type == 'all':
        nstates = mstates = range(Nj)
    else:
        Njcore = core_states(atom.symbol)  # The number of core orbitals
        if type == 'val-val':
            nstates = mstates = range(Njcore, Nj)
        elif type == 'core-core':
            nstates = mstates = range(Njcore)
        elif type == 'val-core':
            nstates = range(Njcore, Nj)
            mstates = range(Njcore)
        else:
            raise RuntimeError('Unknown type of exchange: ', type)

    # Arrays for storing the potential (times radius)
    vr = np.zeros(atom.N)
    vrl = np.zeros(atom.N)

    # do actual calculation of exchange contribution
    Exx = 0.0
    for j1 in nstates:
        # angular momentum of first state
        l1 = atom.l_j[j1]

        for j2 in mstates:
            # angular momentum of second state
            l2 = atom.l_j[j2]

            # joint occupation number
            f12 = 0.5 * (atom.f_j[j1] / (2. * l1 + 1) *
                         atom.f_j[j2] / (2. * l2 + 1))

            # electron density times radius times length element
            nrdr = atom.u_j[j1] * atom.u_j[j2] * atom.dr
            nrdr[1:] /= atom.r[1:]

            # potential times radius
            vr[:] = 0.0

            # L summation
            for l in range(l1 + l2 + 1):
                # get potential for current l-value
                hartree(l, nrdr, atom.r, vrl)

                # take all m1 m2 and m values of Gaunt matrix of the form
                # G(L1,L2,L) where L = {l,m}
                G2 = G_LLL[l1**2:(l1 + 1)**2,
                           l2**2:(l2 + 1)**2,
                           l**2:(l + 1)**2]**2

                # add to total potential
                vr += vrl * np.sum(G2)

            # add to total exchange the contribution from current two states
            Exx += -.5 * f12 * np.dot(vr, nrdr)

    # double energy if mixed contribution
    if type == 'val-core':
        Exx *= 2.

    # return exchange energy
    return Exx


def constructX(gen, gamma=0):
    """Construct the X_p^a matrix for the given atom.

    The X_p^a matrix describes the valence-core interactions of the
    partial waves.
    """
    # initialize attributes
    uv_j = gen.vu_j    # soft valence states * r:
    lv_j = gen.vl_j    # their repective l quantum numbers
    Nvi = 0
    for l in lv_j:
        Nvi += 2 * l + 1   # total number of valence states (including m)

    # number of core and valence orbitals (j only, i.e. not m-number)
    Njcore = gen.njcore
    Njval = len(lv_j)

    # core states * r:
    uc_j = gen.u_j[:Njcore]
    r, dr, N = gen.r, gen.dr, gen.N
    r2 = r**2

    # potential times radius
    vr = np.zeros(N)

    # initialize X_ii matrix
    X_ii = np.zeros((Nvi, Nvi))

    # make gaunt coeff. list
    lmax = max(gen.l_j[:Njcore] + gen.vl_j)
    G_LLL = gaunt(lmax=lmax)

    # sum over core states
    for jc in range(Njcore):
        lc = gen.l_j[jc]

        # sum over first valence state index
        i1 = 0
        for jv1 in range(Njval):
            lv1 = lv_j[jv1]

            # electron density 1 times radius times length element
            n1c = uv_j[jv1] * uc_j[jc] * dr
            n1c[1:] /= r[1:]

            # sum over second valence state index
            i2 = 0
            for jv2 in range(Njval):
                lv2 = lv_j[jv2]

                # electron density 2
                n2c = uv_j[jv2] * uc_j[jc]
                n2c[1:] /= r2[1:]

                # sum expansion in angular momenta
                for l in range(min(lv1, lv2) + lc + 1):
                    # Int density * potential * r^2 * dr:
                    if gamma == 0:
                        vr = gen.rgd.poisson(n2c, l)
                    else:
                        vr = gen.rgd.yukawa(n2c, l, gamma)
                    nv = np.dot(n1c, vr)

                    # expansion coefficients
                    A_mm = X_ii[i1:i1 + 2 * lv1 + 1, i2:i2 + 2 * lv2 + 1]
                    for mc in range(2 * lc + 1):
                        for m in range(2 * l + 1):
                            G1c = G_LLL[lv1**2:(lv1 + 1)**2,
                                        lc**2 + mc, l**2 + m]
                            G2c = G_LLL[lv2**2:(lv2 + 1)**2,
                                        lc**2 + mc, l**2 + m]
                            A_mm += nv * np.outer(G1c, G2c)
                i2 += 2 * lv2 + 1
            i1 += 2 * lv1 + 1

    # pack X_ii matrix
    X_p = pack2(X_ii)
    return X_p


def H_coulomb_val_core(paw, u=0):
    """Short description here.

    ::

                     core   *    *
             //       --   i(r) k(r') k(r) j (r')
       H   = || drdr' >   ----------------------
        ij   //       --          |r - r'|
                      k
    """
    H_nn = np.zeros((paw.wfs.bd.nbands, paw.wfs.bd.nbands),
                    dtype=paw.wfs.dtype)
    for a, P_ni in paw.wfs.kpt_u[u].P_ani.items():
        X_ii = unpack(paw.wfs.setups[a].X_p)
        H_nn += np.dot(P_ni.conj(), np.dot(X_ii, P_ni.T))
    paw.wfs.gd.comm.sum(H_nn)
    from ase.units import Hartree
    return H_nn * Hartree



"""Test the reading of wave functions as file references."""
from math import sqrt
import numpy as np
from ase import Atoms
from gpaw import GPAW
from gpaw.mpi import world, rank


def test_fileio_file_reference(in_tmp_dir):
    d = 3.0
    atoms = Atoms('Na3',
                  positions=[(0, 0, 0),
                             (0, 0, d),
                             (0, d * sqrt(3 / 4), d / 2)],
                  magmoms=[1.0, 1.0, 1.0],
                  cell=(3.5, 3.5, 4 + 2 / 3),
                  pbc=True)

    # Only a short, non-converged calculation
    conv = {'eigenstates': 1.24, 'energy': 2e-1, 'density': 1e-1}
    calc = GPAW(h=0.30, kpts=(1, 1, 3),
                setups={'Na': '1'},
                nbands=3, convergence=conv)
    atoms.calc = calc
    atoms.get_potential_energy()
    wf0 = calc.get_pseudo_wave_function(2, 1, 1)

    calc.write('tmp', 'all')

    # Now read with single process
    comm = world.new_communicator(np.array((rank,)))
    calc = GPAW('tmp', communicator=comm)
    wf1 = calc.get_pseudo_wave_function(2, 1, 1)
    diff = np.abs(wf0 - wf1)
    assert np.all(diff < 1e-12)



import sys
import re

import pytest
import numpy as np
from ase.build import molecule
from ase.units import Hartree

from gpaw import GPAW
from gpaw.mpi import rank, world
from gpaw.test import equal
from gpaw.gauss import Gauss
from gpaw.lrtddft import LrTDDFT, photoabsorption_spectrum
from gpaw.lrtddft.kssingle import KSSingles


@pytest.mark.lrtddft
@pytest.mark.slow
def test_lrtddft_3(in_tmp_dir):
    from io import StringIO

    L = 10.0
    txt = None
    txt = '-'

    N2 = molecule('N2')
    N2.set_cell([L, L, L])
    # N2.set_pbc(True)
    N2.center()

    if 1:
        calc = GPAW(h=0.25,
                    experimental={'niter_fixdensity': 2},
                    nbands=-5,
                    spinpol=True,
                    xc='PBE',
                    txt=txt,
                    eigensolver='cg',
                    parallel={'domain': world.size})
        N2.calc = calc
        _ = N2.get_potential_energy()
        calc.write('N2_wfs.gpw', 'all')

    # selections
    for obj in [KSSingles, LrTDDFT]:
        # selection using state numbers
        el = obj(restrict={'istart': 3, 'jend': 6}, txt=txt)
        el.calculate(N2)
        if hasattr(obj, 'diagonalize'):
            el.diagonalize()
    #    print "*************** obj, len(obj)", obj.__name__, len(el)
        assert len(el) == 8
        # selection using an energy range
        el = obj(restrict={'energy_range': 8}, txt=txt)
        el.calculate(N2)
        if hasattr(obj, 'diagonalize'):
            el.diagonalize()
    #    print "*************** obj, len(obj)", obj.__name__, len(el)
        assert len(el) == 4
        el = obj(restrict={'energy_range': 11.5}, txt=txt)
        el.calculate(N2)
    #    print "*************** obj, len(obj)", obj.__name__, len(el)
        if hasattr(obj, 'diagonalize'):
            el.diagonalize()
        assert len(el) == 18
        if hasattr(obj, 'diagonalize'):
            el.diagonalize(restrict={'energy_range': 8})
            assert len(el) == 4

    lr = LrTDDFT(calc, nspins=2)
    lr.write('lrtddft3.dat.gz')
    lr.diagonalize()

    world.barrier()

    # This is done to test if writing and reading again yields the same result
    lr2 = LrTDDFT.read('lrtddft3.dat.gz')
    lr2.diagonalize()

    # Unfortunately not all of the lrtddft code is parallel
    if rank == 0:
        Epeak = 19.5  # The peak we want to investigate (this is alone)
        Elist = np.asarray([lrsingle.get_energy() * Hartree
                            for lrsingle in lr])
        n = np.argmin(np.abs(Elist - Epeak))  # index of the peak

        E = lr[n].get_energy() * Hartree
        osz = lr[n].get_oscillator_strength()
        print('Original object        :', E, osz[0])

        # Test the output of analyse
        origstdout = sys.stdout
        sys.stdout = sio = StringIO()
        lr.analyse(n)
        s = sio.getvalue()
        sys.stdout = origstdout
        match = re.findall(
            r'%i: E=([0-9]*\.[0-9]*) eV, f=([0-9]*\.[0-9]*)*' % n, s)
        Eanalyse = float(match[0][0])
        oszanalyse = float(match[0][1])
        print('From analyse           :', Eanalyse, oszanalyse)
        equal(E, Eanalyse, 1e-3)            # Written precision in analyse
        equal(osz[0], oszanalyse, 1e-3)

        E2 = lr2[n].get_energy() * Hartree
        osz2 = lr2[n].get_oscillator_strength()
        print('Written and read object:', E2, osz2[0])

        # Compare values of original and written/read objects
        equal(E, E2, 1e-4)
        for i in range(len(osz)):
            equal(osz[i], osz2[i], 1.7e-4)

        width = 0.05
        photoabsorption_spectrum(lr,
                                 spectrum_file='lrtddft3-spectrum.dat',
                                 width=width)
        # We need to be able to check the heights in the spectrum
        weight = Gauss(width).get(0)

        spectrum = np.loadtxt('lrtddft3-spectrum.dat', usecols=(0, 1))
        idx = (spectrum[:, 0] >= E - 0.1) & (spectrum[:, 0] <= E + 0.1)
        peak = np.argmax(spectrum[idx, 1]) + np.nonzero(idx)[0][0]
        Espec = spectrum[peak, 0]
        oszspec = spectrum[peak, 1] / weight

        print('Values from spectrum   :', Espec, oszspec)
        # Compare calculated values with values written to file
        equal(E, Espec, 1e-2)           # The spectrum has a low sampling
        equal(osz[0], oszspec, 1e-2)



import pytest
import numpy as np
from ase.build import bulk
from ase.dft.kpoints import monkhorst_pack
# from ase.units import Bohr

from gpaw import GPAW, FermiDirac, PW
from gpaw.response.chi0 import Chi0
from gpaw.mpi import serial_comm
from itertools import product


@pytest.mark.response
@pytest.mark.slow
def test_response_chi0(in_tmp_dir):
    # inputs to loop over [k, gamma, center, sym]
    settings = product([2, 3], *[[False, True]] * 3)

    for k, gamma, center, sym in settings:
        if k == 3 and gamma:
            continue
        a = bulk('Si', 'diamond')
        q_c = [0, 0, 1.0 / k]
        kpts = monkhorst_pack((k, k, k))
        if gamma:
            kpts += 0.5 / k
        if center:
            a.center()
        name = 'si.k%d.g%d.c%d.s%d' % (k, gamma, center, bool(sym))

        calc = a.calc = GPAW(
            kpts=kpts,
            symmetry={'point_group': sym},
            mode=PW(150),
            occupations=FermiDirac(width=0.001),
            convergence={'bands': 8},
            txt=name + '.txt')
        a.get_potential_energy()
        calc.write(name, 'all')

        calc = GPAW(name, txt=None, communicator=serial_comm)

        chi = Chi0(calc, frequencies=np.array([0, 1.0, 2.0]), hilbert=False,
                   ecut=100, txt=name + '.log')
        chi0 = chi.calculate(q_c)
        assert chi0.blockdist.blockcomm.size == 1
        chi0_wGG = chi0.chi0_WgG  # no block distribution

        # sym and center: False
        if not sym and not center:
            chi00_w = chi0_wGG[:, 0, 0]
        elif -1 not in calc.wfs.kd.bz2bz_ks:
            assert abs(chi0_wGG[:, 0, 0] - chi00_w).max() < 35e-5

        if not sym:
            chi00_wGG = chi0_wGG
        elif -1 not in calc.wfs.kd.bz2bz_ks:
            assert abs(chi0_wGG - chi00_wGG).max() < 2e-5

        chi0 = chi.calculate([0, 0, 0])
        assert chi0.blockdist.blockcomm.size == 1
        chi0_wGG = chi0.chi0_WgG  # no block distribution

        if not sym and not center:
            chi000_w = chi0_wGG[:, 0, 0]
        elif -1 not in calc.wfs.kd.bz2bz_ks:
            assert abs(chi0_wGG[:, 0, 0] - chi000_w).max() < 0.0015

        if not sym:
            chi000_wGG = chi0_wGG
        elif -1 not in calc.wfs.kd.bz2bz_ks:
            assert abs(chi0_wGG - chi000_wGG).max() < 0.0015



import pytest
from gpaw import GPAW
from gpaw.utilities.ibz2bz import ibz2bz


@pytest.mark.legacy
@pytest.mark.serial
def test_ibz2bz(gpw_files, in_tmp_dir):
    """Test ibz.gpw -> bz.gpw utility."""
    path = 'bz.gpw'
    gpw = gpw_files['bcc_li_pw']
    ibz2bz(gpw, path)
    ef1 = GPAW(gpw).get_fermi_level()
    ef2 = GPAW(path).get_fermi_level()
    assert ef1 == ef2



import os
import subprocess
import sys

from ase.utils import import_module
from ase.utils import search_current_git_hash

import gpaw
import _gpaw
import gpaw.fftw as fftw
from gpaw.mpi import rank, have_mpi
from gpaw.utilities import compiled_with_sl, compiled_with_libvdwxc
from gpaw.utilities.elpa import LibElpa


def info():
    """Show versions of GPAW and its dependencies."""
    results = [('python-' + sys.version.split()[0], sys.executable)]
    for name in ['gpaw', 'ase', 'numpy', 'scipy']:
        try:
            module = import_module(name)
        except ImportError:
            results.append((name, False))
        else:
            # Search for git hash
            githash = search_current_git_hash(module)
            if githash is None:
                githash = ''
            else:
                githash = '-{:.10}'.format(githash)
            results.append((name + '-' + module.__version__ + githash,
                            module.__file__.rsplit('/', 1)[0] + '/'))

    libxc = gpaw.libraries['libxc']
    if libxc:
        results.append((f'libxc-{libxc}', True))
    else:
        results.append(('libxc', False))

    module = import_module('_gpaw')
    if hasattr(module, 'githash'):
        githash = '-{:.10}'.format(module.githash())
    else:
        githash = ''
    results.append(('_gpaw' + githash,
                    os.path.normpath(getattr(module, '__file__',
                                             'built-in'))))
    if '_gpaw' in sys.builtin_module_names or not have_mpi:
        p = subprocess.Popen(['which', 'gpaw-python'], stdout=subprocess.PIPE)
        results.append(('parallel',
                        p.communicate()[0].strip().decode() or False))
    results.append(('MPI enabled', have_mpi))
    results.append(('OpenMP enabled', _gpaw.have_openmp))
    results.append(('GPU-aware MPI', getattr(_gpaw, 'gpu_aware_mpi', False)))
    if have_mpi:
        have_sl = compiled_with_sl()
        have_elpa = LibElpa.have_elpa()
        if have_elpa:
            version = LibElpa.api_version()
            if version is None:
                version = 'unknown, at most 2018.xx'
            have_elpa = f'yes; version: {version}'
    else:
        have_sl = have_elpa = 'no (MPI unavailable)'

    if not hasattr(_gpaw, 'mmm'):
        results.append(('BLAS', 'using scipy.linalg.blas and numpy.dot()'))

    results.append(('scalapack', have_sl))
    results.append(('Elpa', have_elpa))

    have_fftw = fftw.have_fftw()
    results.append(('FFTW', have_fftw))
    results.append(('libvdwxc', compiled_with_libvdwxc()))
    for i, path in enumerate(gpaw.setup_paths):
        results.append((f'PAW-datasets ({i + 1})', str(path)))

    if rank != 0:
        return

    lines = [(a, b if isinstance(b, str) else ['no', 'yes'][b])
             for a, b in results]
    n1 = max(len(a) for a, _ in lines)
    n2 = max(len(b) for _, b in lines)
    print(' ' + '-' * (n1 + 4 + n2))
    for a, b in lines:
        print(f'| {a:{n1}}  {b:{n2}} |')
    print(' ' + '-' * (n1 + 4 + n2))


class CLICommand:
    """Show versions of GPAW and its dependencies"""

    @staticmethod
    def add_arguments(parser):
        pass

    @staticmethod
    def run(args):
        info()



"""Code for summerschool notebooks."""
import json
import sys
from pathlib import Path
from typing import Dict, Any, Optional


def py2ipynb(path: Path,
             kernel: Optional[Dict[str, str]] = None,
             teachermode: Optional[bool] = False) -> None:
    """Convert Python script to ipynb file.

    Hides cells marked with "# teacher" and replaces lines marked with
    "# student: ..."::

        answer = 42  # student: answer = ...

    gives::

        answer = ...

    """
    cells = []
    text = path.read_text()
    assert text.startswith('# %%\n')
    chunks = text[5:].split('\n\n# %%\n')

    for chunk in chunks:
        cell_type = 'code'
        if chunk.startswith(('"""', 'r"""')):
            chunk = chunk.strip('r\n')
            chunk = chunk.strip('"')
            cell_type = 'markdown'

        cell: Dict[str, Any] = {
            'cell_type': cell_type,
            'metadata': {},
            'source': chunk.splitlines(True)}

        if cell_type == 'code':
            cell['outputs'] = []
            cell['execution_count'] = None
            lines = cell['source']
            for i, line in enumerate(lines):
                if ' # student:' in line and not teachermode:
                    a, b = (x.strip() for x in line.split('# student:'))
                    lines[i] = line.split(a)[0] + b + '\n'
                elif line.startswith('# magic: '):
                    lines[i] = line[9:]
                elif line.lower().startswith('# teacher') and not teachermode:
                    del lines[i:]
                    break

        cells.append(cell)

    outpath = path.with_suffix('.ipynb')
    if kernel is None:
        kernel = {'display_name': 'Python 3',
                  'language': 'python',
                  'name': 'python3'}
    outpath.write_text(
        json.dumps(
            {'cells': cells,
             'metadata': {
                 'kernelspec': kernel,
                 'language_info': {
                     'codemirror_mode': {'name': 'ipython', 'version': 3},
                     'file_extension': '.py',
                     'mimetype': 'text/x-python',
                     'name': 'python',
                     'nbconvert_exporter': 'python',
                     'pygments_lexer': 'ipython3',
                     'version': '3.6.1'}},
             'nbformat': 4,
             'nbformat_minor': 1},
            indent=2))


if __name__ == '__main__':
    py2ipynb(Path(sys.argv[1]))



import pytest
from gpaw.atom.all_electron import AllElectron
from gpaw.test import equal


@pytest.mark.gllb
@pytest.mark.libxc
def test_gllb_atomic():
    data = []

    def out(a, b, c, d, e, f):
        data.append((a, b, c, d, e, f))

    ETotal = {'Be': -14.572 + 0.012,
              'Ne': -128.548 - 0.029,
              'Mg': -199.612 - 0.005}
    EX = {'Be': -2.666 - 0.010,
          'Ne': -12.107 - 0.122,
          'Mg': -15.992 - 0.092}
    EHOMO = {'Be': -0.309 + 0.008,
             'Ne': -0.851 + 0.098,
             'Mg': -0.253 + 0.006}
    # eignum = {'Be': 0, 'Ne': 3, 'Mg': 0}

    for xcname in ['GLLB', 'GLLBSC']:
        atoms = ['Be', 'Ne', 'Mg']
        for atom in atoms:
            # Test AllElectron GLLB
            GLLB = AllElectron(atom, xcname=xcname, scalarrel=False,
                               gpernode=600)
            GLLB.run()

            out('Total energy', xcname + '1D', atom, ETotal[atom], GLLB.ETotal,
                'Ha')
            out('Exchange energy', xcname + '1D', atom, EX[atom], GLLB.Exc,
                'Ha')
            out('HOMO Eigenvalue', xcname + '1D', atom, EHOMO[atom],
                GLLB.e_j[-1],
                'Ha')
            if xcname == 'GLLB':
                equal(GLLB.ETotal, ETotal[atom], tolerance=1e-2)
                equal(GLLB.Exc, EX[atom], tolerance=1e-2)
                equal(GLLB.e_j[-1], EHOMO[atom], tolerance=1e-2)

        print('             Quanity        Method    Symbol     Ref[1]'
              '         GPAW      Unit  ')
        for a, b, c, d, e, f in data:
            print('%20s %10s %10s   %10.3f   %10.3f   %5s' %
                  (a, b, c, d, e, f))

    print("""References:
    [1] Self-consistent approximation to the Kohn-Sham exchange potential
    Gritsenko, Oleg; Leeuwen, Robert van; Lenthe, Erik van; Baerends, Evert Jan
    Phys. Rev. A Vol. 51 p. 1944""")



import pytest


@pytest.mark.skip(reason='TODO')
def test_generalizedlaue():
    from gpaw.poisson import (NonPeriodicLauePoissonSolver,
                              GeneralizedLauePoissonSolver,
                              FDPoissonSolver, idst2, dst2)
    from gpaw.grid_descriptor import GridDescriptor
    import numpy as np
    gd = GridDescriptor((10, 12, 42), (4, 5, 20), pbc_c=(True, True, False))
    poisson = GeneralizedLauePoissonSolver(nn=2)
    poisson.set_grid_descriptor(gd)

    poisson2 = FDPoissonSolver(nn=2, eps=1e-28)
    poisson2.set_grid_descriptor(gd)

    phi_g = gd.zeros()
    phi2_g = gd.zeros()
    rho_g = gd.zeros()
    rho_g[4, 5, 6] = 1.0
    rho_g[4, 5, 7] = -1.0

    poisson.solve(phi_g, rho_g)
    poisson2.solve(phi2_g, rho_g)
    print("this", phi_g[4, 5, :])
    print("ref", phi2_g[4, 5, :])
    print("diff", phi_g[4, 5, :] - phi2_g[4, 5, :])

    assert np.linalg.norm(phi_g - phi2_g) < 1e-10

    gd = GridDescriptor((10, 12, 42), (10, 12, 42),
                        pbc_c=(False, False, False))
    poisson = NonPeriodicLauePoissonSolver(nn=1)
    poisson.set_grid_descriptor(gd)
    print("eigs", poisson.eigs_c[0])
    poisson2 = FDPoissonSolver(nn=1, eps=1e-24)
    poisson2.set_grid_descriptor(gd)

    phi_g = gd.zeros()
    phi2_g = gd.zeros()
    rho_g = gd.zeros()
    rho_g[:, :, 1] = 1.0
    rho_g[:, :, 20] = -1.0
    poisson.solve(phi_g, rho_g)
    poisson2.solve(phi2_g, rho_g)
    print("this", phi_g[4, 5, :])
    print("ref", phi2_g[4, 5, :])
    print("diff", phi_g[4, 5, :] - phi2_g[4, 5, :])

    assert np.linalg.norm(phi_g - phi2_g) < 1e-10

    X = idst2(dst2(rho_g))
    diff = X - rho_g
    print("dct diff", np.linalg.norm(diff))
    assert np.linalg.norm(diff) < 1e-13



import pytest
import numpy as np
from ase import Atoms
from gpaw import GPAW, FermiDirac, PW
from gpaw.response.df import DielectricFunction
from gpaw.test import equal, findpeak


@pytest.fixture
def gpwfile(in_tmp_dir):
    cluster = Atoms('Au2', [(0, 0, 0), (0, 0, 2.564)])
    cluster.set_cell((6, 6, 6), scale_atoms=False)
    cluster.center()
    calc = GPAW(mode=PW(ecut=180, force_complex_dtype=True),
                xc={'name': 'RPBE', 'stencil': 1},
                nbands=16,
                eigensolver='rmm-diis',
                parallel={'domain': 1},
                occupations=FermiDirac(0.01))

    cluster.calc = calc
    cluster.get_potential_energy()
    calc.diagonalize_full_hamiltonian(nbands=24, scalapack=True)
    gpwname = 'Au2.gpw'
    calc.write(gpwname, 'all')
    return gpwname


@pytest.mark.response
@pytest.mark.slow
def test_response_au02_absorption(scalapack, in_tmp_dir, gpwfile):
    nw = 141
    frequencies = np.linspace(0, 14, nw)

    df = DielectricFunction(gpwfile,
                            frequencies=frequencies,
                            hilbert=False,
                            eta=0.1,
                            ecut=10)

    b0, b = df.get_dielectric_function(filename=None,
                                       direction='z')
    a0, a = df.get_polarizability(filename=None,
                                  direction='z')

    w0_ = 5.60491055
    I0_ = 227.23392824591642
    w_ = 5.644900254787107
    I_ = 184.4086028397282

    w, I = findpeak(frequencies, b0.imag)
    equal(w, w0_, 0.05)
    equal(6**3 * I / (4 * np.pi), I0_, 0.5)
    w, I = findpeak(frequencies, a0.imag)
    equal(w, w0_, 0.05)
    equal(I, I0_, 0.5)
    w, I = findpeak(frequencies, b.imag)
    equal(w, w_, 0.05)
    equal(6**3 * I / (4 * np.pi), I_, 0.5)
    w, I = findpeak(frequencies, a.imag)
    equal(w, w_, 0.05)
    equal(I, I_, 0.5)



import numpy as np

from ase.units import Hartree
from gpaw import GPAW
from gpaw.utilities import unpack
from gpaw.utilities.tools import tri2full, lowdin
from gpaw.lcao.tools import basis_subset2, get_bfi2
from gpaw.coulomb import get_vxc as get_ks_xc
from gpaw.utilities.blas import r2k, mmmx
from gpaw.lcao.projected_wannier import (dots, condition_number, eigvals,
                                         get_bfs, get_lcao_projections_HSP)


def get_rot(F_MM, V_oM, L):
    eps_M, U_MM = np.linalg.eigh(F_MM)
    indices = eps_M.real.argsort()[-L:]
    U_Ml = U_MM[:, indices]
    U_Ml /= np.sqrt(dots(U_Ml.T.conj(), F_MM, U_Ml).diagonal())

    U_ow = V_oM.copy()
    U_lw = np.dot(U_Ml.T.conj(), F_MM)
    for col1, col2 in zip(U_ow.T, U_lw.T):
        norm = np.linalg.norm(np.hstack((col1, col2)))
        col1 /= norm
        col2 /= norm
    return U_ow, U_lw, U_Ml


def get_lcao_xc(calc, P_aqMi, bfs=None, spin=0):
    nq = len(calc.wfs.kd.ibzk_qc)
    nao = calc.wfs.setups.nao
    dtype = calc.wfs.dtype
    if bfs is None:
        bfs = get_bfs(calc)

    if calc.density.nt_sg is None:
        calc.density.interpolate_pseudo_density()
    nt_sg = calc.density.nt_sg
    vxct_sg = calc.density.finegd.zeros(calc.wfs.nspins)
    calc.hamiltonian.xc.calculate(calc.density.finegd, nt_sg, vxct_sg)
    vxct_G = calc.wfs.gd.zeros()
    calc.hamiltonian.restrict_and_collect(vxct_sg[spin], vxct_G)
    Vxc_qMM = np.zeros((nq, nao, nao), dtype)
    for q, Vxc_MM in enumerate(Vxc_qMM):
        bfs.calculate_potential_matrix(vxct_G, Vxc_MM, q)
        tri2full(Vxc_MM, 'L')

    # Add atomic PAW corrections
    for a, P_qMi in P_aqMi.items():
        D_sp = calc.density.D_asp[a][:]
        H_sp = np.zeros_like(D_sp)
        calc.hamiltonian.xc.calculate_paw_correction(calc.wfs.setups[a],
                                                     D_sp, H_sp)
        H_ii = unpack(H_sp[spin])
        for Vxc_MM, P_Mi in zip(Vxc_qMM, P_qMi):
            Vxc_MM += dots(P_Mi, H_ii, P_Mi.T.conj())
    return Vxc_qMM * Hartree


def get_xc2(calc, w_wG, P_awi, spin=0):
    if calc.density.nt_sg is None:
        calc.density.interpolate_pseudo_density()
    nt_g = calc.density.nt_sg[spin]
    vxct_g = calc.density.finegd.zeros()
    calc.hamiltonian.xc.get_energy_and_potential(nt_g, vxct_g)
    vxct_G = calc.wfs.gd.empty()
    calc.hamiltonian.restrict_and_collect(vxct_g, vxct_G)

    # Integrate pseudo part
    Nw = len(w_wG)
    xc_ww = np.empty((Nw, Nw))
    r2k(.5 * calc.wfs.gd.dv, w_wG, vxct_G * w_wG, .0, xc_ww)
    tri2full(xc_ww, 'L')

    # Add atomic PAW corrections
    for a, P_wi in P_awi.items():
        D_sp = calc.density.D_asp[a][:]
        H_sp = np.zeros_like(D_sp)
        calc.wfs.setups[a].xc_correction.calculate_energy_and_derivatives(
            D_sp, H_sp)
        H_ii = unpack(H_sp[spin])
        xc_ww += dots(P_wi, H_ii, P_wi.T.conj())
    return xc_ww * Hartree


class ProjectedWannierFunctionsFBL:
    """PWF in the finite band limit.

    ::

                --N
        |w_w> = >    |psi_n> U_nw
                --n=1
    """
    def __init__(self, V_nM, No, ortho=False):
        Nw = V_nM.shape[1]
        assert No <= Nw

        V_oM, V_uM = V_nM[:No], V_nM[No:]
        F_MM = np.dot(V_uM.T.conj(), V_uM)
        U_ow, U_lw, U_Ml = get_rot(F_MM, V_oM, Nw - No)
        self.U_nw = np.vstack((U_ow, dots(V_uM, U_Ml, U_lw)))

        # stop here ?? XXX
        self.S_ww = self.rotate_matrix(np.ones(1))
        if ortho:
            lowdin(self.U_nw, self.S_ww)
            self.S_ww = np.identity(Nw)
        self.norms_n = np.dot(self.U_nw, np.linalg.solve(
            self.S_ww, self.U_nw.T.conj())).diagonal()

    def rotate_matrix(self, A_nn):
        if A_nn.ndim == 1:
            return np.dot(self.U_nw.T.conj() * A_nn, self.U_nw)
        else:
            return dots(self.U_nw.T.conj(), A_nn, self.U_nw)

    def rotate_projections(self, P_ani):
        P_awi = {}
        for a, P_ni in P_ani.items():
            P_awi[a] = np.tensordot(self.U_nw, P_ni, axes=[[0], [0]])
        return P_awi

    def rotate_function(self, psit_nG):
        return np.tensordot(self.U_nw, psit_nG, axes=[[0], [0]])


class ProjectedWannierFunctionsIBL:
    """PWF in the infinite band limit.

    ::

                --No               --Nw
        |w_w> = >   |psi_o> U_ow + >   |f_M> U_Mw
                --o=1              --M=1
    """
    def __init__(self, V_nM, S_MM, No, lcaoindices=None):
        Nw = V_nM.shape[1]
        assert No <= Nw
        self.V_oM, V_uM = V_nM[:No], V_nM[No:]

        F_MM = S_MM - np.dot(self.V_oM.T.conj(), self.V_oM)
        U_ow, U_lw, U_Ml = get_rot(F_MM, self.V_oM, Nw - No)
        self.U_Mw = np.dot(U_Ml, U_lw)
        self.U_ow = U_ow - np.dot(self.V_oM, self.U_Mw)
        if lcaoindices is not None:
            for i in lcaoindices:
                self.U_ow[:, i] = 0.0
                self.U_Mw[:, i] = 0.0
                self.U_Mw[i, i] = 1.0

        # stop here ?? XXX
        self.S_ww = self.rotate_matrix(np.ones(1), S_MM)
        P_uw = np.dot(V_uM, self.U_Mw)
        self.norms_n = np.hstack((
            np.dot(U_ow,
                   np.linalg.solve(self.S_ww, U_ow.T.conj())).diagonal(),
            np.dot(P_uw,
                   np.linalg.solve(self.S_ww, P_uw.T.conj())).diagonal()))

    def rotate_matrix(self, A_o, A_MM):
        assert A_o.ndim == 1
        A_ww = dots(self.U_ow.T.conj() * A_o, self.V_oM, self.U_Mw)
        A_ww += np.conj(A_ww.T)
        A_ww += np.dot(self.U_ow.T.conj() * A_o, self.U_ow)
        A_ww += dots(self.U_Mw.T.conj(), A_MM, self.U_Mw)
        return A_ww

    def rotate_projections(self, P_aoi, P_aMi, indices=None):
        if indices is None:
            U_ow = self.U_ow
            U_Mw = self.U_Mw
        else:
            U_ow = self.U_ow[:, indices]
            U_Mw = self.U_Mw[:, indices]
        P_awi = {}
        for a, P_oi in P_aoi.items():
            P_awi[a] = np.tensordot(U_Mw, P_aMi[a], axes=[[0], [0]])
            if len(U_ow) > 0:
                P_awi[a] += np.tensordot(U_ow, P_oi, axes=[[0], [0]])
        return P_awi

    def rotate_function(self, psit_oG, bfs, q=-1, indices=None):
        if indices is None:
            U_ow = self.U_ow
            U_Mw = self.U_Mw
        else:
            U_ow = self.U_ow[:, indices]
            U_Mw = self.U_Mw[:, indices]
        w_wG = np.zeros((U_ow.shape[1],) + psit_oG.shape[1:])
        if len(U_ow) > 0:
            mmmx(1, U_ow.T.copy(), 'N', psit_oG, 'N', 0, w_wG)
        bfs.lcao_to_grid(U_Mw.T.copy(), w_wG, q)
        return w_wG


class PWFplusLCAO(ProjectedWannierFunctionsIBL):
    def __init__(self, V_nM, S_MM, No, pwfmask, lcaoindices=None):
        Nw = V_nM.shape[1]
        self.V_oM = V_nM[:No]
        dtype = V_nM.dtype

        # Do PWF optimization for pwfbasis submatrix only!
        Npwf = len(pwfmask.nonzero()[0])
        pwfmask2 = np.outer(pwfmask, pwfmask)
        s_MM = S_MM[pwfmask2].reshape(Npwf, Npwf)
        v_oM = self.V_oM[:, pwfmask]
        f_MM = s_MM - np.dot(v_oM.T.conj(), v_oM)
        nw = len(s_MM)
        assert No <= nw

        u_ow, u_lw, u_Ml = get_rot(f_MM, v_oM, nw - No)
        u_Mw = np.dot(u_Ml, u_lw)
        u_ow = u_ow - np.dot(v_oM, u_Mw)

        # Determine U for full lcao basis
        self.U_ow = np.zeros((No, Nw), dtype)
        for U_w, u_w in zip(self.U_ow, u_ow):
            np.place(U_w, pwfmask, u_w)
        self.U_Mw = np.identity(Nw, dtype)
        np.place(self.U_Mw, pwfmask2, u_Mw.flat)

        if lcaoindices is not None:
            for i in lcaoindices:
                self.U_ow[:, i] = 0.0
                self.U_Mw[:, i] = 0.0
                self.U_Mw[i, i] = 1.0

        self.S_ww = self.rotate_matrix(np.ones(1), S_MM)
        self.norms_n = None


class PWF2:
    def __init__(self, gpwfilename, fixedenergy=0., spin=0, ibl=True,
                 basis='sz', zero_fermi=False, pwfbasis=None, lcaoatoms=None,
                 projection_data=None):
        calc = GPAW(gpwfilename, txt=None, basis=basis)
        assert calc.wfs.gd.comm.size == 1
        assert calc.wfs.kd.comm.size == 1
        assert calc.wfs.bd.comm.size == 1
        if zero_fermi:
            try:
                Ef = calc.get_fermi_level()
            except NotImplementedError:
                Ef = calc.get_homo_lumo().mean()
        else:
            Ef = 0.0

        self.ibzk_kc = calc.get_ibz_k_points()
        self.nk = len(self.ibzk_kc)
        self.eps_kn = [calc.get_eigenvalues(kpt=q, spin=spin) - Ef
                       for q in range(self.nk)]
        self.M_k = [sum(eps_n <= fixedenergy) for eps_n in self.eps_kn]
        print('Fixed states:', self.M_k)
        self.calc = calc
        self.dtype = self.calc.wfs.dtype
        self.spin = spin
        self.ibl = ibl
        self.pwf_q = []
        self.norms_qn = []
        self.S_qww = []
        self.H_qww = []

        if ibl:
            if pwfbasis is not None:
                pwfmask = basis_subset2(calc.atoms.get_chemical_symbols(),
                                        basis, pwfbasis)
            if lcaoatoms is not None:
                lcaoindices = get_bfi2(calc.atoms.get_chemical_symbols(),
                                       basis,
                                       lcaoatoms)
            else:
                lcaoindices = None
            self.bfs = get_bfs(calc)
            if projection_data is None:
                V_qnM, H_qMM, S_qMM, self.P_aqMi = get_lcao_projections_HSP(
                    calc, bfs=self.bfs, spin=spin, projectionsonly=False)
            else:
                V_qnM, H_qMM, S_qMM, self.P_aqMi = projection_data
            H_qMM -= Ef * S_qMM
            for q, M in enumerate(self.M_k):
                if pwfbasis is None:
                    pwf = ProjectedWannierFunctionsIBL(V_qnM[q], S_qMM[q], M,
                                                       lcaoindices)
                else:
                    pwf = PWFplusLCAO(V_qnM[q], S_qMM[q], M, pwfmask,
                                      lcaoindices)
                self.pwf_q.append(pwf)
                self.norms_qn.append(pwf.norms_n)
                self.S_qww.append(pwf.S_ww)
                self.H_qww.append(pwf.rotate_matrix(self.eps_kn[q][:M],
                                                    H_qMM[q]))
        else:
            if projection_data is None:
                V_qnM = get_lcao_projections_HSP(calc, spin=spin)
            else:
                V_qnM = projection_data
            for q, M in enumerate(self.M_k):
                pwf = ProjectedWannierFunctionsFBL(V_qnM[q], M, ortho=False)
                self.pwf_q.append(pwf)
                self.norms_qn.append(pwf.norms_n)
                self.S_qww.append(pwf.S_ww)
                self.H_qww.append(pwf.rotate_matrix(self.eps_kn[q]))

        for S in self.S_qww:
            print('Condition number: %0.1e' % condition_number(S))

    def get_hamiltonian(self, q=0, indices=None):
        if indices is None:
            return self.H_qww[q]
        else:
            return self.H_qww[q].take(indices, 0).take(indices, 1)

    def get_overlap(self, q=0, indices=None):
        if indices is None:
            return self.S_qww[q]
        else:
            return self.S_qww[q].take(indices, 0).take(indices, 1)

    def get_projections(self, q=0, indices=None):
        kpt = self.calc.wfs.kpt_qs[q][self.spin]
        if not hasattr(self, 'P_awi'):
            if self.ibl:
                M = self.M_k[q]
                self.P_awi = self.pwf_q[q].rotate_projections(
                    dict([(a, P_ni[:M]) for a, P_ni in kpt.P_ani.items()]),
                    dict([(a, P_qMi[q]) for a, P_qMi in self.P_aqMi.items()]),
                    indices)
            else:
                assert 0
                # XXX pwf?
                # self.P_awi = pwf.rotate_projections(kpt.P_ani, indices)
        return self.P_awi

    def get_orbitals(self, q=0, indices=None):
        self.calc.wfs.initialize_wave_functions_from_restart_file()
        kpt = self.calc.wfs.kpt_qs[q][self.spin]
        if not hasattr(self, 'w_wG'):
            if self.ibl:
                self.w_wG = self.pwf_q[q].rotate_function(
                    kpt.psit_nG[:self.M_k[q]], self.bfs, q, indices)
            else:
                self.w_wG = self.pwf_q[q].rotate_function(
                    kpt.psit_nG, indices)
        return self.w_wG

    def get_Fcore(self, q=0, indices=None):
        if indices is None:
            Fcore_ww = np.zeros_like(self.H_qww[q])
        else:
            Fcore_ww = np.zeros((len(indices), len(indices)))
        for a, P_wi in self.get_projections(q, indices).items():
            X_ii = unpack(self.calc.wfs.setups[a].X_p)
            Fcore_ww -= dots(P_wi, X_ii, P_wi.T.conj())
        return Fcore_ww * Hartree

    def get_eigs(self, q=0):
        return eigvals(self.H_qww[q], self.S_ww[q])

    def get_condition_number(self, q=0):
        return condition_number(self.S_qww[q])

    def get_xc(self, q=0, indices=None):
        # self.calc.density.ghat.set_positions(
        #     self.calc.spos_ac)
        # self.calc.hamiltonian.poisson.initialize()
        if self.ibl:
            return get_xc2(self.calc, self.get_orbitals(q, indices),
                           self.get_projections(q, indices), self.spin)
        else:
            return self.pwf_q[q].rotate_matrix(get_ks_xc(self.calc,
                                                         spin=self.spin))


class LCAOwrap:
    def __init__(self, calc, spin=0):
        assert calc.wfs.gd.comm.size == 1
        assert calc.wfs.kd.comm.size == 1
        assert calc.wfs.bd.comm.size == 1

        from gpaw.lcao.tools import get_lcao_hamiltonian
        H_skMM, S_kMM = get_lcao_hamiltonian(calc)

        self.calc = calc
        self.dtype = calc.wfs.dtype
        self.spin = spin
        self.H_qww = H_skMM[spin]
        self.S_qww = S_kMM
        self.P_aqwi = calc.wfs.P_aqMi
        self.Nw = self.S_qww.shape[-1]

        for S in self.S_qww:
            print('Condition number: %0.1e' % condition_number(S))

    def get_hamiltonian(self, q=0, indices=None):
        if indices is None:
            return self.H_qww[q]
        else:
            return self.H_qww[q].take(indices, 0).take(indices, 1)

    def get_overlap(self, q=0, indices=None):
        if indices is None:
            return self.S_qww[q]
        else:
            return self.S_qww[q].take(indices, 0).take(indices, 1)

    def get_projections(self, q=0, indices=None):
        if indices is None:
            return dict([(a, P_qwi[q]) for a, P_qwi in self.P_aqwi.items()])
        else:
            return dict([(a, P_qwi[q].take(indices, 0))
                         for a, P_qwi in self.P_aqwi.items()])

    def get_orbitals(self, q=-1, indices=None):
        assert q == -1
        if indices is None:
            indices = range(self.Nw)
        Ni = len(indices)
        C_wM = np.zeros((Ni, self.Nw), self.dtype)
        for i, C_M in zip(indices, C_wM):
            C_M[i] = 1.0
        w_wG = self.calc.wfs.gd.zeros(Ni, dtype=self.dtype)
        self.calc.wfs.basis_functions.lcao_to_grid(C_wM, w_wG, q=-1)
        return w_wG

    def get_Fcore(self, q=0, indices=None):
        if indices is None:
            Fcore_ww = np.zeros_like(self.H_qww[q])
        else:
            Fcore_ww = np.zeros((len(indices), len(indices)))
        for a, P_wi in self.get_projections(q, indices).items():
            if self.calc.wfs.setups[a].type != 'ghost':
                X_ii = unpack(self.calc.wfs.setups[a].X_p)
                Fcore_ww -= dots(P_wi, X_ii, P_wi.T.conj())
        return Fcore_ww * Hartree

    def get_xc(self, q=0, indices=None):
        if not hasattr(self, 'Vxc_qww'):
            self.Vxc_qww = get_lcao_xc(self.calc, self.P_aqwi,
                                       bfs=self.calc.wfs.basis_functions,
                                       spin=self.spin)
        if indices is None:
            return self.Vxc_qww[q]
        else:
            return self.Vxc_qww[q].take(indices, 0).take(indices, 1)



from gpaw.mixer import DummyMixer
from gpaw.xc import XC

from gpaw.lcaotddft.laser import create_laser
from gpaw.lcaotddft.utilities import read_uMM
from gpaw.lcaotddft.utilities import read_wuMM
from gpaw.lcaotddft.utilities import write_uMM
from gpaw.lcaotddft.utilities import write_wuMM


class TimeDependentPotential(object):
    def __init__(self):
        self.ext_i = []
        self.laser_i = []
        self.initialized = False

    def add(self, ext, laser):
        self.ext_i.append(ext)
        self.laser_i.append(laser)

    def initialize(self, paw):
        if self.initialized:
            return

        self.ksl = paw.wfs.ksl
        self.kd = paw.wfs.kd
        self.kpt_u = paw.wfs.kpt_u
        get_matrix = paw.wfs.eigensolver.calculate_hamiltonian_matrix
        self.V_iuMM = []
        for ext in self.ext_i:
            V_uMM = []
            hamiltonian = KickHamiltonian(paw.hamiltonian, paw.density, ext)
            for kpt in paw.wfs.kpt_u:
                V_MM = get_matrix(hamiltonian, paw.wfs, kpt,
                                  add_kinetic=False, root=-1)
                V_uMM.append(V_MM)
            self.V_iuMM.append(V_uMM)
        self.Ni = len(self.ext_i)
        self.initialized = True

    def get_MM(self, u, time):
        V_MM = self.laser_i[0].strength(time) * self.V_iuMM[0][u]
        for i in range(1, self.Ni):
            V_MM += self.laser_i[i].strength(time) * self.V_iuMM[i][u]
        return V_MM

    def write(self, writer):
        writer.write(Ni=self.Ni)
        writer.write(laser_i=[laser.todict() for laser in self.laser_i])
        write_wuMM(self.kd, self.ksl, writer, 'V_iuMM',
                   self.V_iuMM, range(self.Ni))

    def read(self, reader):
        self.Ni = reader.Ni
        self.laser_i = [create_laser(**laser) for laser in reader.laser_i]
        self.V_iuMM = read_wuMM(self.kpt_u, self.ksl, reader, 'V_iuMM',
                                range(self.Ni))
        self.initialized = True


class KickHamiltonian(object):
    def __init__(self, ham, dens, ext):
        nspins = ham.nspins
        vext_g = ext.get_potential(ham.finegd)
        vext_G = ham.restrict_and_collect(vext_g)
        self.vt_sG = ham.gd.empty(nspins)
        for s in range(nspins):
            self.vt_sG[s] = vext_G

        dH_asp = ham.setups.empty_atomic_matrix(nspins,
                                                ham.atom_partition)
        W_aL = dens.ghat.dict()
        dens.ghat.integrate(vext_g, W_aL)
        # XXX this is a quick hack to get the distribution right
        # It'd be better to have these distributions abstracted elsewhere.
        # The idea is (thanks Ask, see discussion in !910):
        # * gd has D cores and coarse grid
        # * aux_gd is the same grid as gd but with either D or
        #   K * B * D cores (with augment_grids = True)
        # * finegd is then aux_gd.refine().
        # In integrals like W_aL, the atomic quantities are distributed
        # according to the domains of those atoms (so finegd for W_aL,
        # but gd for dH_asp).
        # And, some things (atomic XC corrections) are calculated evenly
        # distributed among all cores.
        dHaux_asp = ham.atomdist.to_aux(dH_asp)
        for a, W_L in W_aL.items():
            setup = dens.setups[a]
            dH_p = setup.Delta_pL @ W_L
            for s in range(nspins):
                dHaux_asp[a][s] = dH_p
        self.dH_asp = ham.atomdist.from_aux(dHaux_asp)


class TimeDependentHamiltonian(object):
    def __init__(self, fxc=None, td_potential=None, scale=None,
                 rremission=None):
        assert fxc is None or isinstance(fxc, str)
        self.fxc_name = fxc
        if isinstance(td_potential, dict):
            td_potential = [td_potential]
        if isinstance(td_potential, list):
            pot_i = td_potential
            td_potential = TimeDependentPotential()
            for pot in pot_i:
                td_potential.add(**pot)
        self.td_potential = td_potential
        self.has_scale = scale is not None
        if self.has_scale:
            self.scale = scale
        self.rremission = rremission

    def write(self, writer):
        if self.has_fxc:
            self.write_fxc(writer.child('fxc'))
        if self.has_scale:
            self.write_scale(writer.child('scale'))
        if self.td_potential is not None:
            self.td_potential.write(writer.child('td_potential'))

    def write_fxc(self, writer):
        writer.write(name=self.fxc_name)
        write_uMM(self.wfs.kd, self.wfs.ksl, writer, 'deltaXC_H_uMM',
                  self.deltaXC_H_uMM)

    def write_scale(self, writer):
        writer.write(scale=self.scale)
        write_uMM(self.wfs.kd, self.wfs.ksl, writer, 'scale_H_uMM',
                  self.scale_H_uMM)

    def read(self, reader):
        if 'fxc' in reader:
            self.read_fxc(reader.fxc)
        if 'scale' in reader:
            self.read_scale(reader.scale)
        if 'td_potential' in reader:
            assert self.td_potential is None
            self.td_potential = TimeDependentPotential()
            self.td_potential.ksl = self.wfs.ksl
            self.td_potential.kd = self.wfs.kd
            self.td_potential.kpt_u = self.wfs.kpt_u
            self.td_potential.read(reader.td_potential)

    def read_fxc(self, reader):
        assert self.fxc_name is None or self.fxc_name == reader.name
        self.fxc_name = reader.name
        self.deltaXC_H_uMM = read_uMM(self.wfs.kpt_u, self.wfs.ksl, reader,
                                      'deltaXC_H_uMM')

    def read_scale(self, reader):
        assert not self.has_scale or self.scale == reader.scale
        self.has_scale = True
        self.scale = reader.scale
        self.scale_H_uMM = read_uMM(self.wfs.kpt_u, self.wfs.ksl, reader,
                                    'scale_H_uMM')

    def initialize(self, paw):
        self.timer = paw.timer
        self.timer.start('Initialize TDDFT Hamiltonian')
        self.wfs = paw.wfs
        self.density = paw.density
        self.hamiltonian = paw.hamiltonian
        niter = paw.niter
        if self.rremission is not None:
            self.rremission.initialize(paw)

        # Reset the density mixer
        # XXX: density mixer is not written to the gpw file
        # XXX: so we need to set it always
        self.density.set_mixer(DummyMixer())
        self.update()

        # Initialize fxc
        self.initialize_fxc(niter)

        # Initialize scale
        self.initialize_scale(niter)

        # Initialize td_potential
        if self.td_potential is not None:
            self.td_potential.initialize(paw)

        self.timer.stop('Initialize TDDFT Hamiltonian')

    def initialize_fxc(self, niter):
        self.has_fxc = self.fxc_name is not None
        if not self.has_fxc:
            return
        self.timer.start('Initialize fxc')
        # XXX: Similar functionality is available in
        # paw.py: PAW.linearize_to_xc(self, newxc)
        # See test/lcaotddft/fxc_vs_linearize.py

        # Calculate deltaXC: 1. take current H_MM
        if niter == 0:
            def get_H_MM(kpt):
                return self.get_hamiltonian_matrix(kpt, time=0.0,
                                                   addfxc=False, addpot=False,
                                                   scale=False)
            self.deltaXC_H_uMM = []
            for kpt in self.wfs.kpt_u:
                self.deltaXC_H_uMM.append(get_H_MM(kpt))

        # Update hamiltonian.xc
        if self.fxc_name == 'RPA':
            xc_name = 'null'
        else:
            xc_name = self.fxc_name
        # XXX: xc is not written to the gpw file
        # XXX: so we need to set it always
        xc = XC(xc_name)
        xc.initialize(self.density, self.hamiltonian, self.wfs)
        xc.set_positions(self.hamiltonian.spos_ac)
        self.hamiltonian.xc = xc
        self.update()

        # Calculate deltaXC: 2. update with new H_MM
        if niter == 0:
            for u, kpt in enumerate(self.wfs.kpt_u):
                self.deltaXC_H_uMM[u] -= get_H_MM(kpt)
        self.timer.stop('Initialize fxc')

    def initialize_scale(self, niter):
        if not self.has_scale:
            return
        self.timer.start('Initialize scale')

        # Take current H_MM and multiply with scale
        if niter == 0:
            def get_H_MM(kpt):
                return self.get_hamiltonian_matrix(kpt, time=0.0,
                                                   addfxc=False, addpot=False,
                                                   scale=False)
            self.scale_H_uMM = []
            for kpt in self.wfs.kpt_u:
                self.scale_H_uMM.append((1 - self.scale) * get_H_MM(kpt))
        self.timer.stop('Initialize scale')

    def update_projectors(self):
        self.timer.start('Update projectors')
        for kpt in self.wfs.kpt_u:
            self.wfs.atomic_correction.calculate_projections(self.wfs, kpt)
        self.timer.stop('Update projectors')

    def get_hamiltonian_matrix(self, kpt, time, addfxc=True, addpot=True,
                               scale=True):
        self.timer.start('Calculate H_MM')
        kpt_rank, q = self.wfs.kd.get_rank_and_index(kpt.k)
        u = q * self.wfs.nspins + kpt.s
        assert kpt_rank == self.wfs.kd.comm.rank

        get_matrix = self.wfs.eigensolver.calculate_hamiltonian_matrix
        H_MM = get_matrix(self.hamiltonian, self.wfs, kpt, root=-1)
        if addfxc and self.has_fxc:
            H_MM += self.deltaXC_H_uMM[u]

        if self.rremission is not None:
            H_MM += self.rremission.vradiationreaction(kpt, time)

        if scale and self.has_scale:
            H_MM *= self.scale
            H_MM += self.scale_H_uMM[u]
        if addpot and self.td_potential is not None:
            H_MM += self.td_potential.get_MM(u, time)
        self.timer.stop('Calculate H_MM')
        return H_MM

    def update(self, mode='all'):
        self.timer.start('Update TDDFT Hamiltonian')
        if mode in ['all', 'density']:
            self.update_projectors()
            self.density.update(self.wfs)
        if mode in ['all']:
            self.hamiltonian.update(self.density)
        self.timer.stop('Update TDDFT Hamiltonian')
